#!/usr/bin/env python3

import json
import os
import re
import tempfile
import time
import unittest

from contextlib import contextmanager
from shutil import copytree, which

from selenium import webdriver
from selenium.common.exceptions import (
    InvalidArgumentException,
    NoSuchWindowException,
    TimeoutException,
    WebDriverException,
)
from selenium.webdriver.chrome.options import Options as ChromeOptions
from selenium.webdriver.edge.options import Options as EdgeOptions
from selenium.webdriver.firefox.options import Options as FirefoxOptions
from selenium.webdriver.firefox.service import Service as FirefoxService
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.common.by import By

try:
    from xvfbwrapper import Xvfb
except ImportError:
    print("\n\nxvfbwrapper Python package import failed")
    print("headless mode (ENABLE_XVFB=1) is not supported")


SEL_DEFAULT_WAIT_TIMEOUT = 30

BROWSER_TYPES = ['chrome', 'firefox', 'edge']
BROWSER_NAMES = ['google-chrome', 'google-chrome-stable', 'google-chrome-beta',
                 'firefox', 'microsoft-edge', 'microsoft-edge-beta']


class WindowNotFoundException(Exception):
    pass


def get_browser_type(string):
    for t in BROWSER_TYPES:
        if t in string.lower():
            return t
    raise ValueError(f"Could not get browser type from {string}")


def get_browser_name(string):
    for bn in BROWSER_NAMES:
        if string in bn:
            bn_path = which(bn)
            if bn_path:
                return os.path.basename(bn_path)
    raise ValueError(f"Could not get browser name from {string}")


class Shim:
    _browser_msg = '''BROWSER should be one of:
* /path/to/a/browser
* a browser executable name so we can find the browser with "which $BROWSER"
* something from BROWSER_TYPES
'''
    __doc__ = 'Chooses the correct driver and extension_url based on the BROWSER environment\nvariable. ' + _browser_msg

    def __init__(self):
        print("\n\nConfiguring the test run ...")

        browser = os.environ.get('BROWSER')

        # get browser_path and browser_type first
        if browser is None:
            raise ValueError("The BROWSER environment variable is not set. " + self._browser_msg)

        if ("/" in browser) or ("\\" in browser):
            # path to a browser binary
            self.browser_path = browser
            self.browser_type = get_browser_type(browser)
        elif which(browser):
            # executable browser name like 'google-chrome-stable'
            self.browser_path = which(browser)
            self.browser_type = get_browser_type(browser)
        elif get_browser_type(browser):
            # browser type like 'firefox' or 'chrome'
            bname = get_browser_name(browser)
            self.browser_path = which(bname)
            self.browser_type = get_browser_type(browser)
        else:
            raise ValueError(f"Could not infer BROWSER from {browser}")

        self.extension_path = os.path.abspath(
            os.path.join(os.path.dirname(__file__), '..', '..', 'src'))

        if self.browser_type in ('chrome', 'edge'):
            # this extension ID and the "key" property in manifest.json
            # must both be derived from the same private key
            self.info = {
                'extension_id': 'mcgekeccgjgcmhnhbabplanchdogjcnh'
            }
            self.base_url = f"chrome-extension://{self.info['extension_id']}/"
            if self.browser_type == 'chrome':
                self.manager = self.chrome_manager
            else:
                self.manager = self.edge_manager

            # make extension ID constant across runs
            self.fix_chrome_extension_id()

        elif self.browser_type == 'firefox':
            self.info = {
                'extension_id': 'jid1-MnnxcxisBPnSXQ@jetpack',
                'uuid': 'd56a5b99-51b6-4e83-ab23-796216679614'
            }
            self.manager = self.firefox_manager
            self.base_url = f"moz-extension://{self.info['uuid']}/"

        print(f"\nUsing browser path: {self.browser_path}\n"
              f"with browser type: {self.browser_type}\n"
              f"and extension path: {self.extension_path}\n")

    def fix_chrome_extension_id(self):
        # create temp directory
        # pylint: disable-next=consider-using-with
        self.tmp_dir = tempfile.TemporaryDirectory()
        new_extension_path = os.path.join(self.tmp_dir.name, "src")

        # copy extension sources there
        copytree(self.extension_path, new_extension_path)

        # update manifest.json
        manifest_path = os.path.join(new_extension_path, "manifest.json")
        with open(manifest_path, "r", encoding="utf-8") as f:
            manifest = json.load(f)
        # this key and the extension ID must both be derived from the same private key
        manifest['key'] = "MIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEArMdgFkGsm7nOBr/9qkx8XEcmYSu1VkIXXK94oXLz1VKGB0o2MN+mXL/Dsllgkh61LZgK/gVuFFk89e/d6Vlsp9IpKLANuHgyS98FKx1+3sUoMujue+hyxulEGxXXJKXhk0kGxWdE0IDOamFYpF7Yk0K8Myd/JW1U2XOoOqJRZ7HR6is1W6iO/4IIL2/j3MUioVqu5ClT78+fE/Fn9b/DfzdX7RxMNza9UTiY+JCtkRTmm4ci4wtU1lxHuVmWiaS45xLbHphQr3fpemDlyTmaVoE59qG5SZZzvl6rwDah06dH01YGSzUF1ezM2IvY9ee1nMSHEadQRQ2sNduNZWC9gwIDAQAB" # noqa:E501 pylint:disable=line-too-long
        with open(manifest_path, "w", encoding="utf-8") as f:
            json.dump(manifest, f)

        # update self.extension_path
        self.extension_path = new_extension_path

    @property
    def wants_xvfb(self):
        if self.on_github_actions or bool(int(os.environ.get('ENABLE_XVFB', 0))):
            try:
                Xvfb
            except NameError:
                print("\nHeadless mode not supported: install xvfbwrapper first")
                return False
            return True
        return False

    @property
    def on_github_actions(self):
        if "GITHUB_ACTIONS" in os.environ:
            return True
        return False

    @contextmanager
    def chrome_manager(self):
        opts = ChromeOptions()

        # support tests in Chrome with MV2 PB while this flag remains available
        opts.add_argument("--enable-features=AllowLegacyMV2Extensions")

        opts.binary_location = self.browser_path
        opts.enable_bidi = True
        opts.enable_webextensions = True

        # work around https://issues.chromium.org/issues/409441960
        opts.add_experimental_option('enableExtensionTargets', True)

        # https://github.com/GoogleChromeLabs/chromium-bidi/issues/3281
        opts.set_capability("unhandledPromptBehavior", "ignore");

        # TODO not yet in Firefox (w/o hacks anyway):
        # https://github.com/mozilla/geckodriver/issues/284#issuecomment-456073771
        opts.set_capability("goog:loggingPrefs", {'browser': 'ALL'})

        for i in range(5):
            try:
                driver = webdriver.Chrome(options=opts)
                driver.webextension.install(self.extension_path)
            except WebDriverException as ex:
                if i == 0: print("")
                print(f"ChromeDriver initialization failed: {ex}")
            else:
                break

        try:
            yield driver
        finally:
            driver.quit()

    @contextmanager
    def edge_manager(self):
        opts = EdgeOptions()

        opts.binary_location = self.browser_path
        opts.enable_bidi = True
        opts.enable_webextensions = True

        # work around https://issues.chromium.org/issues/409441960
        opts.add_experimental_option('enableExtensionTargets', True)

        # https://github.com/GoogleChromeLabs/chromium-bidi/issues/3281
        opts.set_capability("unhandledPromptBehavior", "ignore");

        for i in range(5):
            try:
                driver = webdriver.Edge(options=opts)
                driver.webextension.install(self.extension_path)
            except WebDriverException as ex:
                if i == 0: print("")
                print(f"EdgeDriver initialization failed ({i+1}/5): {ex}", end='')
            else:
                break

        try:
            yield driver
        finally:
            driver.quit()

    @contextmanager
    def firefox_manager(self):
        opts = FirefoxOptions()

        opts.binary_location = self.browser_path
        opts.enable_bidi = True
        opts.enable_webextensions = True

        # https://github.com/mozilla/geckodriver/issues/2241#issuecomment-3984861843
        opts.set_capability("unhandledPromptBehavior", "ignore");

        # make extension ID constant across runs
        opts.set_preference('extensions.webextensions.uuids', '{"%s": "%s"}' % (
            self.info['extension_id'], self.info['uuid']))

        # needed for test_referrer_header()
        # https://bugzilla.mozilla.org/show_bug.cgi?id=1720294
        opts.set_preference('network.http.referer.disallowCrossSiteRelaxingDefault', False)

        # disable tracker cookie blocking as it breaks cookie tests
        # that use trackersimulator.org, a "known tracker",
        # and disable cookie site isolation, as it breaks the cookie
        # tracking detection test
        opts.set_preference("network.cookie.cookieBehavior", 0)

        # disable JSON viewer as it breaks parsing JSON pages
        opts.set_preference("devtools.jsonview.enabled", False)

        # to produce a trace-level geckodriver.log,
        # set the log_output argument to FirefoxService() to "./geckodriver.log"
        # and uncomment the line below
        #opts.log.level = "trace"

        for i in range(5):
            try:
                service = FirefoxService(log_output=os.path.devnull)
                driver = webdriver.Firefox(options=opts, service=service)
                driver.webextension.install(self.extension_path)
            except WebDriverException as ex:
                if i == 0: print("")
                print(f"Firefox WebDriver initialization failed ({i+1}/5): {ex}", end='')
            else:
                break

        try:
            yield driver
        finally:
            driver.quit()


shim = Shim() # create the browser shim


def retry_until(fun, tester=None, times=3, msg=None):
    """
    Execute function `fun` until either its return is truthy
    (or if `tester` is set, until the result of calling `tester` with `fun`'s return is truthy),
    or it gets executed X times, where X = `times` + 1.
    """
    for i in range(times):
        result = fun()

        if tester is not None:
            if tester(result):
                break
        elif result:
            break

        if msg:
            if i == 0:
                print("")
            print(msg)

        time.sleep(2 ** i)

    return result


class PBSeleniumTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.manager = shim.manager
        cls.base_url = shim.base_url
        cls.wants_xvfb = shim.wants_xvfb
        if cls.wants_xvfb:
            cls.vdisplay = Xvfb(width=1280, height=720)
            cls.vdisplay.start()

    @classmethod
    def tearDownClass(cls):
        if cls.wants_xvfb:
            cls.vdisplay.stop()

    def init(self, driver):
        self.driver = driver
        self.js = self.driver.execute_script
        self.options_url = self.base_url + "skin/options.html"
        self.popup_url = self.base_url + "skin/popup.html"
        self.first_run_url = self.base_url + "skin/firstRun.html"
        self.test_url = self.base_url + "tests/index.html"

    def run(self, result=None):
        # disable Selenium sending your IP to a third-party company
        # https://www.selenium.dev/documentation/selenium_manager/#data-collection
        os.environ['SE_AVOID_STATS'] = "true"

        with self.manager() as driver:
            self.init(driver)

            # wait for Badger's storage, listeners, ...
            self.load_url(self.options_url)
            self.wait_for_script(
                "let done = arguments[arguments.length - 1];"
                "chrome.runtime.sendMessage({"
                "  type: 'isBadgerInitialized'"
                "}, r => done(r));", execute_async=True)

            # also disable the welcome page
            self.driver.execute_async_script(
                "let done = arguments[arguments.length - 1];"
                "chrome.runtime.sendMessage({"
                "  type: 'updateSettings',"
                "  data: { showIntroPage: false }"
                "}, () => {"
                "   chrome.tabs.query({}, (res) => {"
                "     let welcome_tab = res && res.find("
                "       tab => tab.url == chrome.runtime.getURL('skin/firstRun.html'));"
                "     if (!welcome_tab) {"
                "       return done();"
                "     }"
                "     chrome.tabs.remove(welcome_tab.id, done);"
                "   });"
                "});")

            super().run(result)

    def is_firefox_nightly(self):
        caps = self.driver.capabilities
        if caps['browserName'] == "firefox":
            version = caps['browserVersion']
            return re.search('a[0-9]+$', version) is not None
        return False

    def open_window(self):
        try:
            self.driver.switch_to.new_window('tab')
        except NoSuchWindowException:
            time.sleep(1)
            self.driver.switch_to.new_window('tab')

    def load_url(self, url, wait_for_body_text=False, retries=5):
        """Load a URL and wait before returning."""
        for i in range(retries):
            try:
                self.driver.get(url)
                break
            except TimeoutException as e:
                if i < retries - 1:
                    time.sleep(2 ** i)
                    continue
                raise e
            # work around geckodriver/marionette/Firefox timeout handling,
            # for example: https://travis-ci.org/EFForg/privacybadger/jobs/389429089
            except WebDriverException as e:
                if str(e).startswith("Reached error page") and i < retries - 1:
                    time.sleep(2 ** i)
                    continue
                raise e

        if wait_for_body_text:
            # wait for document.body.textContent to become truthy
            retry_until(
                lambda: self.driver.find_element(By.TAG_NAME, 'body').text)

    def txt_by_css(self, css_selector, timeout=SEL_DEFAULT_WAIT_TIMEOUT):
        """Find an element by CSS selector and return its text."""
        return self.find_el_by_css(
            css_selector, visible_only=False, timeout=timeout).text

    def find_el_by_css(self, css_selector, visible_only=True, timeout=SEL_DEFAULT_WAIT_TIMEOUT):
        condition = (
            EC.visibility_of_element_located if visible_only
            else EC.presence_of_element_located
        )
        return WebDriverWait(self.driver, timeout).until(
            condition((By.CSS_SELECTOR, css_selector)))

    def find_el_by_xpath(self, xpath, timeout=SEL_DEFAULT_WAIT_TIMEOUT):
        return WebDriverWait(self.driver, timeout).until(
            EC.visibility_of_element_located((By.XPATH, xpath)))

    @contextmanager
    def wait_for_reload(self, timeout=SEL_DEFAULT_WAIT_TIMEOUT):
        """Context manager that waits for the page to reload,
        to be used with actions that reload the page."""
        page = self.driver.find_element(By.TAG_NAME, 'html')
        yield
        try:
            WebDriverWait(self.driver, timeout).until(EC.staleness_of(page))
        except WebDriverException as e:
            # work around Firefox nonsense
            if str(e).startswith("Message: TypeError: can't access dead object"):
                pass
            else:
                raise e

    def wait_for_script(self, script, *script_args,
        timeout=SEL_DEFAULT_WAIT_TIMEOUT, execute_async=False,
        message="Timed out waiting for execute_script to eval to True"):

        """Variant of execute_script/execute_async_script
        that keeps rerunning the script until it returns True."""

        def execute_script(dr):
            if execute_async: return dr.execute_async_script(script, *script_args)
            return dr.execute_script(script, *script_args)

        return WebDriverWait(self.driver, timeout).until(execute_script, message)

    def wait_for_any_text(self, selector, timeout=SEL_DEFAULT_WAIT_TIMEOUT):
        return WebDriverWait(self.driver, timeout).until(
            lambda d: d.find_element(By.CSS_SELECTOR, selector).text.strip())

    def wait_for_text(self, selector, text, timeout=SEL_DEFAULT_WAIT_TIMEOUT):
        return WebDriverWait(self.driver, timeout).until(
            EC.text_to_be_present_in_element(
                (By.CSS_SELECTOR, selector), text))

    def wait_for_and_switch_to_frame(self, selector, timeout=SEL_DEFAULT_WAIT_TIMEOUT):
        return WebDriverWait(self.driver, timeout).until(
            EC.frame_to_be_available_and_switch_to_it(
                (By.CSS_SELECTOR, selector)))

    def switch_to_window_with_url(self, url, max_tries=5):
        """Point the driver to the first window that matches this url."""

        for _ in range(max_tries):
            for w in self.driver.window_handles:
                try:
                    self.driver.switch_to.window(w)
                    if self.driver.current_url != url:
                        continue
                except (InvalidArgumentException, NoSuchWindowException):
                    pass
                except WebDriverException as e:
                    if "cannot determine loading status" in str(e):
                        pass
                    else:
                        raise e
                else:
                    return

            time.sleep(1)

        raise WindowNotFoundException("Failed to find window for " + url)


    def close_window_with_url(self, url, max_tries=5):
        self.switch_to_window_with_url(url, max_tries)

        if len(self.driver.window_handles) == 1:
            # open another window to avoid implicit session deletion
            self.open_window()
            self.switch_to_window_with_url(url, max_tries)

        self.driver.close()
        self.driver.switch_to.window(self.driver.window_handles[0])
        self.open_window()

    def set_dnt(self, domain):
        self.load_url(self.options_url)
        self.driver.execute_async_script(
            "let done = arguments[arguments.length - 1];"
            "chrome.runtime.sendMessage({"
            "  type: 'setDnt',"
            "  domain: arguments[0]"
            "}, done);", domain)

    def check_dnt(self, domain):
        self.load_url(self.options_url)
        return self.driver.execute_async_script(
            "let done = arguments[arguments.length - 1];"
            "chrome.runtime.sendMessage({"
            "  type: 'checkForDntPolicy',"
            "  domain: arguments[0]"
            "}, done);", domain)

    def set_user_action(self, domain, action):
        """Adds or modifies the action_map entry for `domain`,
        setting userAction to "user_" + `action`."""
        self.load_url(self.options_url)
        self.driver.execute_async_script(
            "let done = arguments[arguments.length - 1];"
            "chrome.runtime.sendMessage({"
            "  type: 'saveOptionsToggle',"
            "  domain: arguments[0],"
            "  action: arguments[1]"
            "}, done);", domain, action)

    def add_domain(self, domain, action):
        """Adds or modifies the action_map entry for `domain`,
        setting heuristicAction to `action`."""
        self.load_url(self.options_url)
        self.driver.execute_async_script(
            "let done = arguments[arguments.length - 1],"
            "  domain = arguments[0],"
            "  action = arguments[1];"
            "chrome.runtime.sendMessage({"
            "  type: 'setAction', domain, action"
            "}, done);", domain, action)

    def block_domain(self, domain):
        self.add_domain(domain, "block")

    def cookieblock_domain(self, domain):
        self.add_domain(domain, "cookieblock")

    def add_site_override(self, domain, site_domain):
        self.load_url(self.options_url)
        self.driver.execute_async_script(
            "let done = arguments[arguments.length - 1];"
            "chrome.runtime.sendMessage({"
            "  type: 'addSiteOverride',"
            "  domain: arguments[0],"
            "  site_domain: arguments[1]"
            "}, done);", domain, site_domain)

    def disable_badger_on_site(self, url):
        self.load_url(self.options_url)
        self.wait_for_script("return window.OPTIONS_INITIALIZED")
        self.driver.find_element(By.ID, 'new-disabled-site-input').send_keys(url)
        self.driver.find_element(By.CSS_SELECTOR, '#add-disabled-site').click()

    def reenable_badger_on_site(self, domain):
        self.load_url(self.options_url)
        self.driver.execute_async_script(
            "let done = arguments[arguments.length - 1];"
            "chrome.runtime.sendMessage({"
            "  type: 'reenableOnSites',"
            "  domains: [arguments[0]]"
            "}, done);", domain)

    def get_domain_slider_state(self, domain):
        label = self.driver.find_element(
            By.CSS_SELECTOR, f'input[name="{domain}"][checked]')
        return label.get_property('value')

    def clear_tracker_data(self):
        self.load_url(self.options_url)
        self.driver.execute_async_script(
            "let done = arguments[arguments.length - 1];"
            "chrome.runtime.sendMessage({"
            "  type: 'removeAllData'"
            "}, done);")

    def get_badger_storage(self, store_name):
        self.load_url(self.options_url)
        return self.driver.execute_async_script((
            "let done = arguments[arguments.length - 1],"
            "  store_name = arguments[0];"
            "chrome.runtime.sendMessage({"
            "  type: 'syncStorage',"
            "  storeName: store_name"
            "}, function () {"
            "  chrome.storage.local.get([store_name], function (res) {"
            "    done(res[store_name]);"
            "  });"
            "});"
        ), store_name)

    def open_popup(self, target_url=None, show_reminder=False):
        """Show the PB popup as a new tab.

        If Selenium would let us just programmatically launch an extension from its icon,
        we wouldn't need this method. Alas it will not.

        But! We can open a new tab and set the url to the extension's popup html page and
        test away. That's how most devs test extensions. But**2!! PB's popup code uses
        the current tab's url to report the current tracker status.  And since we changed
        the current tab's url when we loaded the popup as a tab, the popup loses all the
        blocker status information from the original tab.

        The workaround is to execute a new convenience function in the popup codebase that
        looks for a given url in the tabs and, if it finds a match, refreshes the popup
        with the associated tabid. Then the correct status information will be displayed
        in the popup."""

        for _ in range(5):
            self.open_window()
            self.load_url(self.popup_url)
            self.wait_for_script("return window.POPUP_INITIALIZED")

            # get the popup populated with status information for the correct url
            self.js("""
/**
 * @param {String} [url]
 * @param {Boolean} [show_reminder]
 */
(function (url, show_reminder) {
  let queryOpts = { currentWindow: true };
  if (url) {
    queryOpts = { url };
  }
  chrome.tabs.query(queryOpts, function (tabs) {
    if (!tabs || !tabs.length) {
      return;
    }
    chrome.runtime.sendMessage({
      type: "getPopupData",
      tabId: tabs[0].id,
      tabUrl: tabs[0].url
    }, (response) => {
      response.settings.seenComic = !show_reminder;
      setPopupData(response);
      refreshPopup();
      showNagMaybe(); // not init() because init() already ran and should only run once
      window.DONE_REFRESHING = true;
    });
  });
}(arguments[0], arguments[1]));""", target_url, show_reminder)

            try:
                # wait for popup to be ready
                self.wait_for_script("return window.DONE_REFRESHING && window.SLIDERS_DONE")
            except TimeoutException:
                continue
            else:
                break

    def get_tracker_state(self):
        """Parse the UI to group all third party domains into their respective action states."""

        notYetBlocked = {}
        cookieBlocked = {}
        blocked = {}

        domain_divs = self.driver.find_elements(By.CSS_SELECTOR,
            "#blockedResourcesInner > div.clicker[data-origin]")
        for div in domain_divs:
            domain = div.get_dom_attribute('data-origin')

            # assert that this domain is never duplicated in the UI
            self.assertNotIn(domain, notYetBlocked)
            self.assertNotIn(domain, cookieBlocked)
            self.assertNotIn(domain, blocked)

            # get slider state for given domain
            action_type = self.get_domain_slider_state(domain)

            # non-tracking domains are hidden by default
            # so if we see a slider set to "allow",
            # it must be in the tracking-but-not-yet-blocked section
            if action_type == 'allow':
                notYetBlocked[domain] = True
            elif action_type == 'cookieblock':
                cookieBlocked[domain] = True
            elif action_type == 'block':
                blocked[domain] = True
            else:
                self.fail(f"what is this?!? {action_type}")

        return {
            'notYetBlocked': notYetBlocked,
            'cookieBlocked': cookieBlocked,
            'blocked': blocked
        }

    @property
    def logs(self):
        # TODO not yet in Firefox
        return [log.get('message') for log in self.driver.get_log('browser')]



#!/usr/bin/env python

import unittest

import pbtest

from selenium.common.exceptions import TimeoutException
from selenium.webdriver.common.by import By


class QUnitTest(pbtest.PBSeleniumTest):

    def test_run_qunit_tests(self):
        self.load_url(self.test_url)

        try:
            # this text appears when tests finish running
            self.txt_by_css(
                "#qunit-testresult-display > span.total",
                timeout=120
            )
        except TimeoutException as exc:
            self.fail(f"Cannot find the results of QUnit tests {exc}")

        print("\nQUnit summary:")
        print(self.txt_by_css("#qunit-testresult-display"))

        failed_test_els = self.driver.find_elements(By.CSS_SELECTOR,
            ".fail .test-name")
        fail_msg = "The following QUnit tests failed:\n  * {}".format(
            "\n  * ".join([el.text for el in failed_test_els])
        )

        assert not failed_test_els, fail_msg


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python

import unittest

import pytest

import pbtest


class ClobberingTest(pbtest.PBSeleniumTest):
    def test_localstorage_clobbering(self):
        LOCALSTORAGE_TESTS = [
            # (test result element ID, expected stored, expected empty)
            ('get-item', "qwerty", "null"),
            ('get-property', "asdf", "undefined"),
            ('get-item-proto', "qwerty", "null"),
            ('get-item-srcdoc', "qwerty", "null"),
            ('get-property-srcdoc', "asdf", "undefined"),
            ('get-item-frames', "qwerty", "null"),
            ('get-property-frames', "asdf", "undefined"),
        ]
        # page loads a frame that writes to and reads from localStorage
        # TODO remove delays from fixture once race condition (https://crbug.com/478183) is fixed
        FIXTURE_URL = "https://privacybadger-tests.eff.org/html/clobbering.html"
        FRAME_DOMAIN = "efforg.github.io"

        # first allow localStorage to be set
        self.load_url(FIXTURE_URL)
        self.wait_for_and_switch_to_frame('iframe')
        for selector, expected, _ in LOCALSTORAGE_TESTS:
            # wait for each test to run
            self.wait_for_script(
                "return document.getElementById('%s')"
                ".textContent != '...';" % selector,
                timeout=2,
                message=(
                    "Timed out waiting for localStorage (%s) to finish ... "
                    "This probably means the fixture "
                    "errored out somewhere." % selector
                )
            )
            self.assertEqual(
                expected, self.txt_by_css("#" + selector),
                "localStorage (%s) was not read successfully"
                "for some reason" % selector
            )

        # mark the frame domain for cookieblocking
        self.cookieblock_domain(FRAME_DOMAIN)

        # now rerun and check results for various localStorage access tests
        self.load_url(FIXTURE_URL)
        self.wait_for_and_switch_to_frame('iframe')
        for selector, _, expected in LOCALSTORAGE_TESTS:
            # wait for each test to run
            self.wait_for_script(
                "return document.getElementById('%s')"
                ".textContent != '...';" % selector,
                timeout=2,
                message=(
                    "Timed out waiting for localStorage (%s) to finish ... "
                    "This probably means the fixture "
                    "errored out somewhere." % selector
                )
            )
            self.assertEqual(
                expected, self.txt_by_css("#" + selector),
                "localStorage (%s) was read despite cookieblocking" % selector
            )

    @pytest.mark.flaky(reruns=9)
    def test_referrer_header(self):
        FIXTURE_URL = (
            "https://efforg.github.io/privacybadger-test-fixtures/html/"
            "referrer.html"
        )
        THIRD_PARTY_DOMAIN = "httpbingo.org"

        def verify_referrer_header(expected, failure_message):
            self.load_url(FIXTURE_URL)
            self.wait_for_script(
                "return document.getElementById('referrer').textContent != '';",
                timeout=10)
            referrer = self.txt_by_css("#referrer")
            self.assertEqual("Referer=", referrer[0:8], "Unexpected page output")
            self.assertEqual(expected, referrer[8:], failure_message)

        # verify base case
        verify_referrer_header(
            FIXTURE_URL,
            "Unexpected default referrer header")

        # cookieblock the domain fetched by the fixture
        self.cookieblock_domain(THIRD_PARTY_DOMAIN)

        # recheck what the referrer header looks like now after cookieblocking
        verify_referrer_header(
            "https://efforg.github.io/",
            "Referrer header does not appear to be origin-only")


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3

import json
import sys

KEYS = set([
    'action_map',
    'fp_scripts',
    'snitch_map',
    'tracking_map',
    'version',
])

with open(sys.argv[1], encoding="utf-8") as f:
    try:
        js = json.load(f)
        if set(js.keys()) == KEYS:
            sys.exit(0)
        else:
            print(f"json keys {js.keys()} are not correct")
            sys.exit(1)
    except Exception as e:
        print("error parsing json:", e)
        sys.exit(1)



#!/usr/bin/env python

import unittest

import pbtest

from functools import partial


class FingerprintingTest(pbtest.PBSeleniumTest):
    """Tests to make sure fingerprinting detection works as expected."""

    def get_fillText_source(self):
        return self.js("""
            const canvas = document.getElementById("writetome");
            const ctx = canvas.getContext("2d");
            return ctx.fillText.toString();
        """)

    def setUp(self):
        # enable local learning
        self.load_url(self.options_url)
        self.wait_for_script("return window.OPTIONS_INITIALIZED")
        self.find_el_by_css('a[href="#tab-general-settings"]').click()
        self.find_el_by_css('#local-learning-checkbox').click()

    def test_canvas_fingerprinting_detection(self, frozen_error=False):
        SITE_DOMAIN = "efforg.github.io"
        FIXTURE = "fingerprinting_frozen_error.html" if frozen_error else "fingerprinting.html"
        FIXTURE_URL = f"https://{SITE_DOMAIN}/privacybadger-test-fixtures/html/{FIXTURE}"

        FP_DOMAIN = "cdn.jsdelivr.net"
        FP_BASE_DOMAIN = "jsdelivr.net"
        FP_PATH = "/npm/fingerprintjs2@1.5.1/dist/fingerprint2.min.js"

        # clear pre-trained/seed tracker data
        self.clear_tracker_data()

        # visit the page
        self.load_url(FIXTURE_URL)

        # open popup and check slider state
        def get_sliders(url, category):
            self.open_popup(url)
            sliders = self.get_tracker_state()
            return sliders[category]
        sliders = pbtest.retry_until(
            partial(get_sliders, FIXTURE_URL, 'notYetBlocked'), times=3)
        assert FP_DOMAIN in sliders, (
            "Canvas fingerprinting domain should be reported in the popup")

        # check that we detected canvas fingerprinting specifically
        assert "canvas" in self.get_badger_storage('tracking_map')\
            .get(FP_BASE_DOMAIN, {}).get(SITE_DOMAIN, []), (
                "Failed to detect canvas fingerprinting script")

        # confirm we recorded the script's domain and path combo
        assert FP_PATH in self.get_badger_storage('fp_scripts').get(FP_DOMAIN, {}), (
            "Failed to record the fingerprinting script's path")

    def test_canvas_fingerprinting_detection_frozen(self):
        self.test_canvas_fingerprinting_detection(True)

    # Privacy Badger overrides a few functions on canvas contexts to check for fingerprinting.
    # In previous versions, it would restore the native function after a single call. Unfortunately,
    # this would wipe out polyfills that had also overridden the same functions, such as the very
    # popular "hidpi-canvas-polyfill".
    def test_canvas_polyfill_clobbering(self):
        FIXTURE_URL = (
            "https://efforg.github.io/privacybadger-test-fixtures/html/"
            "fingerprinting_canvas_hidpi.html"
        )

        # visit the page
        self.load_url(FIXTURE_URL)

        # check that we did not restore the native function (should be hipdi polyfill)
        assert "[native code]" not in self.get_fillText_source(), (
            "Canvas context fillText is not native version (polyfill has been retained)")


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3

import json
import sys

from collections import OrderedDict

def convert(text):
    patterns = [
        "https://docs.google.com/*",
        "http://docs.google.com/*",
        "https://mail.google.com/*",
        "http://mail.google.com/*",
    ]
    for domain in text.split():
        patterns.append("https://www" + domain + "/*")
        patterns.append("http://www" + domain + "/*")
    return patterns

def update_manifest(tempfile_path, manifest_path):
    with open(manifest_path, 'r', encoding="utf-8") as f:
        manifest = json.load(f, object_pairs_hook=OrderedDict)

    with open(tempfile_path, 'r+', encoding="utf-8") as f:
        # tempfile_path contains Google's supported domains
        match_patterns = convert(f.read())

        scripts_idx = -1
        for idx, entry in enumerate(manifest['content_scripts']):
            if "js/firstparties/google.js" in entry['js']:
                scripts_idx = idx
                break
        if scripts_idx == -1:
            print("Failed to locate the Google Search content script in the manifest!")
            sys.exit(1)

        manifest['content_scripts'][scripts_idx]['matches'] = match_patterns

        # overwrite tempfile_path with the updated manifest
        f.seek(0)
        # print() auto-adds a trailing newline
        print(
            json.dumps(
                manifest,
                sort_keys=False,
                indent=2,
                separators=(',', ': ')
            ),
            file=f
        )
        f.truncate()


if __name__ == '__main__':
    # argv[1]: the path to a copy of https://www.google.com/supported_domains
    # argv[2]: the path to the extension manifest
    update_manifest(sys.argv[1], sys.argv[2])



#!/usr/bin/env python3

import json
import subprocess
import sys

from collections import OrderedDict


def prune_fp_scripts(data):
    export_js = """
// shim just enough for constants.js to load
globalThis.chrome = { runtime: { getURL: ()=>{} } };
const { default: constants } = await import('./src/js/constants.js');
process.stdout.write(JSON.stringify(Array.from(constants.FP_CDN_DOMAINS)));"""

    try:
        cmd = ["node", "--experimental-default-type=module", f'--eval={export_js}']
        fp_cdn_domains = subprocess.run(
            cmd, capture_output=True, check=True, text=True).stdout.strip()
    except subprocess.CalledProcessError as ex:
        print(ex.stderr, file=sys.stderr)
        raise ex

    new_fp_scripts = {}
    for domain in data['fp_scripts']:
        if domain in fp_cdn_domains:
            new_fp_scripts[domain] = data['fp_scripts'][domain]

    data['fp_scripts'] = new_fp_scripts

    return data


def prune_action_map(data):
    new_action_map = {}

    for domain in data['action_map']:
        keep = False
        action_entry = data['action_map'][domain]

        # keep base domains,
        # and DNT-compliant or cookieblocked subdomains
        if domain in data['snitch_map']:
            keep = True
        else:
            if 'dnt' in action_entry:
                keep = True
            elif action_entry['heuristicAction'] == "cookieblock":
                keep = True

        if keep:
            new_action_map[domain] = action_entry

    data['action_map'] = new_action_map

    return data


if __name__ == "__main__":
    if len(sys.argv) != 2:
        print(f"Usage: {sys.argv[0]} BADGER_SEED_DATA.json")
        sys.exit(1)

    with open(sys.argv[1], 'r+', encoding='utf-8') as seed_file:
        # read in seed data, preserving ordering
        seed_data = json.load(seed_file, object_pairs_hook=OrderedDict)

        seed_data = prune_action_map(seed_data)

        seed_data = prune_fp_scripts(seed_data)

        # write the data back out
        seed_file.seek(0)
        seed_file.truncate(0)
        # this should match how data gets written out by Badger Sett
        json.dump(seed_data, seed_file, indent=2, sort_keys=True, separators=(',', ': '))



#!/usr/bin/env python

import unittest

import pbtest

PB_POLICY_HASH_LEN = 40  # https://www.eff.org/files/dnt-policies.json


class StorageTest(pbtest.PBSeleniumTest):
    """Privacy Badger storage initialization tests."""

    def test_should_init_storage_entries(self):
        policy_hashes = self.get_badger_storage('dnt_hashes')
        for policy_hash in policy_hashes.keys():
            assert PB_POLICY_HASH_LEN == len(policy_hash)

        badger_settings = self.get_badger_storage('settings_map')
        assert badger_settings['disabledSites'] == [], (
            "Shouldn't have any disabledSites after installation")
        assert badger_settings['checkForDNTPolicy'], (
            "Should start with DNT policy enabled")

        private_settings = self.get_badger_storage('private_storage')
        version_in_storage = private_settings['badgerVersion']
        version_in_manifest = self.js("return chrome.runtime.getManifest().version")
        assert version_in_storage == version_in_manifest, (
            "private_storage should contain the correct version string")


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3

import json

from collections import OrderedDict
from glob import glob


def minify(locale):
    # read in locale, preserving existing ordering
    with open(locale, 'r', encoding="utf-8") as f:
        data = json.load(f, object_pairs_hook=OrderedDict)

    # blank out descriptions
    for key in data:
        if "description" in data[key]:
            data[key]["description"] = ""

    with open(locale, 'w', encoding="utf-8") as f:
        json.dump(data, f, ensure_ascii=False, indent=4)


if __name__ == '__main__':
    for locale in glob("src/_locales/*/*.json"):
        minify(locale)



#!/usr/bin/env python

import unittest

import pbtest

from selenium.common.exceptions import TimeoutException

from pbtest import retry_until


class SurrogatesTest(pbtest.PBSeleniumTest):
    """Integration tests to verify surrogate script functionality."""

    FIXTURE_HOST = "efforg.github.io"
    FIXTURE_URL = (
        f"https://{FIXTURE_HOST}/privacybadger-test-fixtures/html/"
        "ga_surrogate.html"
    )
    SURROGATE_HOST_BASE = "google-analytics.com"
    SURROGATE_HOST = f"www.{SURROGATE_HOST_BASE}"

    def load_ga_js_fixture(self, cache_workaround=False, timeout=12):
        self.load_url(self.FIXTURE_URL)

        if cache_workaround and self.is_firefox_nightly():
            self.driver.refresh()

        load_status_sel = '#third-party-load-result'
        self.wait_for_script(
            "return document.querySelector(arguments[0]).textContent",
            load_status_sel, timeout=timeout)
        if self.find_el_by_css(load_status_sel).text == "error":
            return False

        try:
            self.wait_for_and_switch_to_frame('iframe', timeout=timeout)
            self.wait_for_text('h1', "It worked!", timeout=timeout)
            self.driver.switch_to.default_content()
            return True
        except TimeoutException:
            return False

    def test_ga_js_surrogate(self):
        # clear pre-trained/seed tracker data
        self.clear_tracker_data()

        # verify site loads
        assert self.load_ga_js_fixture(), (
            "page failed to load even before we did anything")

        # block ga.js (should break the site)
        self.block_domain(self.SURROGATE_HOST)
        # disable surrogates
        self.driver.execute_async_script(
            "let done = arguments[arguments.length - 1];"
            "chrome.runtime.sendMessage({"
            "  type: 'disableSurrogates'"
            "}, done);")

        # verify site breaks
        assert not self.load_ga_js_fixture(True), (
            "page loaded successfully when it should have failed")

        # re-enable surrogates
        self.load_url(self.options_url)
        self.driver.execute_async_script(
            "let done = arguments[arguments.length - 1];"
            "chrome.runtime.sendMessage({"
            "  type: 'restoreSurrogates'"
            "}, done);")

        # verify site loads again
        assert retry_until(self.load_ga_js_fixture), (
            "page failed to load after surrogation")

    def test_cookieblocking_base_overwrites_subdomain_surrogate(self):
        SURROGATE_TOSTRING = "function() {\n    }"

        def get_tracker_tostring():
            return self.js("return _gat._getTrackers.toString();")

        self.block_domain(self.SURROGATE_HOST)

        assert self.load_ga_js_fixture()
        # verify we replaced SURROGATE_HOST with surrogate
        assert get_tracker_tostring() == SURROGATE_TOSTRING, (
            "tracker does not appear to have been replaced with surrogate")

        self.cookieblock_domain(self.SURROGATE_HOST_BASE)

        assert self.load_ga_js_fixture()
        # verify that we are no longer replacing SURROGATE_HOST with surrogate
        assert get_tracker_tostring() != SURROGATE_TOSTRING, (
            "surrogation took place when it shouldn't have")

    def test_userblock_applies_surrogate(self):
        self.block_domain(self.SURROGATE_HOST)
        self.set_user_action(self.SURROGATE_HOST, "block")
        assert retry_until(self.load_ga_js_fixture), (
            "page widget unexpectedly failed")
        # also verify that surrogation took place
        text_sel = "#script-string"
        self.wait_for_any_text(text_sel)
        assert self.find_el_by_css(text_sel).text == "function() { }", (
            "tracker does not appear to have been replaced with surrogate")

    def test_site_override_for_domain_with_surrogate(self):
        text_sel = "#script-string"

        self.block_domain(self.SURROGATE_HOST)
        self.add_site_override(self.SURROGATE_HOST, self.FIXTURE_HOST)

        # assert the actual tracker loaded, not the surrogate
        self.load_url(self.FIXTURE_URL)
        self.wait_for_any_text(text_sel)
        assert self.find_el_by_css(text_sel).text != "function() { }", (
            "still redirecting to surrogate despite site-specific override")


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python

import time
import unittest

import pbtest

from selenium.common.exceptions import TimeoutException
from selenium.webdriver.common.by import By
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import WebDriverWait


class PopupTest(pbtest.PBSeleniumTest):
    """Make sure the popup works correctly."""

    def wait_for_sliders(self, visible=True, timeout=5):
        def invisibility_condition(dr):
            """tests all are invisible"""
            sliders = dr.find_elements(By.CSS_SELECTOR, "div.clicker")
            for slider in sliders:
                if slider.is_displayed():
                    return False
            return True

        try:
            if visible:
                WebDriverWait(self.driver, timeout).until(
                    EC.visibility_of_all_elements_located((By.CSS_SELECTOR, "div.clicker")))
            else:
                WebDriverWait(self.driver, timeout).until(
                    invisibility_condition)
        except TimeoutException:
            pass

    def wait_for_page_to_start_loading(self, url, timeout=20):
        """Wait until the title element is present. Use it to work around
        Firefox not updating self.driver.current_url fast enough."""
        try:
            WebDriverWait(self.driver, timeout).until(
                EC.presence_of_element_located((By.CSS_SELECTOR, "title")))
        except TimeoutException:
            # TODO debug info
            print("\n")
            print("driver.current_url=" + self.driver.current_url)
            print()
            print(self.driver.page_source[:5000])
            print("...\n")

            self.fail(f"Timed out waiting for {url} to start loading")

    def test_welcome_page_reminder_overlay(self):
        """Ensure overlay links to new user welcome page."""

        # first close the welcome page if already open
        try:
            self.close_window_with_url(self.first_run_url, max_tries=1)
        except pbtest.WindowNotFoundException:
            pass

        self.open_popup(show_reminder=True)
        self.driver.find_element(By.ID, "intro-reminder-btn").click()

        # switch to the welcome page or fail
        self.switch_to_window_with_url(self.first_run_url)

    def test_help_button(self):
        """Ensure FAQ website is opened when help button is clicked."""

        FAQ_URL = "https://privacybadger.org/#faq"

        try:
            self.switch_to_window_with_url(FAQ_URL, max_tries=1)
        except pbtest.WindowNotFoundException:
            pass
        else:
            self.fail("FAQ should not already be open")

        self.open_popup()
        self.driver.find_element(By.ID, "help").click()

        self.switch_to_window_with_url(FAQ_URL)

    def test_options_button(self):
        """Ensure options page is opened when button is clicked."""
        # first close the options page if already open
        try:
            self.switch_to_window_with_url(self.options_url, max_tries=1)
            self.open_window()
            self.switch_to_window_with_url(self.options_url)
            self.driver.close()
            self.driver.switch_to.window(self.driver.window_handles[0])
        except pbtest.WindowNotFoundException:
            pass
        self.open_popup()
        self.driver.find_element(By.ID, "options").click()
        self.switch_to_window_with_url(self.options_url)

    def test_trackers_link(self):
        """Ensure trackers link opens EFF website."""

        EFF_URL = "https://privacybadger.org/#What-is-a-third-party-tracker"
        DUMMY_PAGE_URL = "https://privacybadger-tests.eff.org/html/widget_frame.html"

        self.load_url(DUMMY_PAGE_URL)
        self.open_popup(DUMMY_PAGE_URL)
        self.driver.find_element(By.CSS_SELECTOR, "#instructions-no-trackers a").click()

        # Make sure EFF website not opened in same window.
        if self.driver.current_url != self.popup_url:
            self.fail("EFF website not opened in new window")

        # Look for EFF website and return if found.
        self.switch_to_window_with_url(EFF_URL)

        self.wait_for_page_to_start_loading(EFF_URL)

        assert self.driver.current_url == EFF_URL, (
            "EFF website should open after clicking trackers link on popup")

        # Verify EFF website contains the linked anchor element.
        faq_selector = f"a[href='{EFF_URL[EFF_URL.index('#'):]}']"
        try:
            WebDriverWait(self.driver, pbtest.SEL_DEFAULT_WAIT_TIMEOUT).until(
                EC.presence_of_element_located((By.CSS_SELECTOR, faq_selector)))
        except TimeoutException:
            self.fail(f"Unable to find expected element ({faq_selector}) on EFF website")

    def test_toggling_sliders(self):
        """Ensure toggling sliders is persisted."""

        FIXTURE_URL = ("https://privacybadger-tests.eff.org/html/"
            "assorted_thirdparties.html")
        DOMAIN = "efforg.github.io"
        DOMAIN_ID = DOMAIN.replace(".", "-")

        self.clear_tracker_data()

        # enable showing non-tracking domains in the popup
        self.wait_for_script("return window.OPTIONS_INITIALIZED")
        self.find_el_by_css('a[href="#tab-general-settings"]').click()
        self.find_el_by_css('#local-learning-checkbox').click()
        self.find_el_by_css('#show-nontracking-domains-checkbox').click()

        self.load_url(FIXTURE_URL)
        self.open_popup(FIXTURE_URL)

        # click input with JavaScript to avoid "Element ... is not clickable" /
        # "Other element would receive the click" Selenium limitation
        self.js(f"$('#block-{DOMAIN_ID}').click()")

        # retrieve the new action
        self.load_url(self.options_url)
        self.wait_for_script("return window.OPTIONS_INITIALIZED")
        self.find_el_by_css('a[href="#tab-tracking-domains"]').click()
        assert self.get_domain_slider_state(DOMAIN) == "block", (
            "The domain should be blocked on options page")

        self.open_popup(FIXTURE_URL)
        assert self.driver.find_element(By.ID, "block-" + DOMAIN_ID).is_selected(), (
            "The domain should be shown as blocked in popup")

        # change to "cookieblock"
        self.js(f"$('#cookieblock-{DOMAIN_ID}').click()")
        # change again to "block"
        self.js(f"$('#block-{DOMAIN_ID}').click()")

        # retrieve the new action
        self.load_url(self.options_url)
        self.wait_for_script("return window.OPTIONS_INITIALIZED")
        self.find_el_by_css('a[href="#tab-tracking-domains"]').click()
        assert self.get_domain_slider_state(DOMAIN) == "block", (
            "The domain should still be blocked on options page")

    def test_reverting_control(self):
        """Test restoring control of a domain to Privacy Badger."""

        FIXTURE_URL = ("https://privacybadger-tests.eff.org/html/"
            "assorted_thirdparties.html")
        DOMAIN = "efforg.github.io"
        DOMAIN_ID = DOMAIN.replace(".", "-")

        self.clear_tracker_data()

        # record the domain as cookieblocked by Badger
        self.cookieblock_domain(DOMAIN)

        self.load_url(FIXTURE_URL)
        self.open_popup(FIXTURE_URL)

        # reveal sliders
        self.driver.find_element(By.ID, 'expand-blocked-resources').click()
        self.wait_for_sliders()

        # set the domain to user control
        # click input with JavaScript to avoid "Element ... is not clickable" /
        # "Other element would receive the click" Selenium limitation
        self.js(f"$('#block-{DOMAIN_ID}').click()")

        self.load_url(self.options_url)
        self.wait_for_script("return window.OPTIONS_INITIALIZED")
        self.find_el_by_css('a[href="#tab-tracking-domains"]').click()
        action = self.get_domain_slider_state(DOMAIN)
        assert action == "block", "Domain should be marked as blocked"
        assert self.driver.find_element(By.CSS_SELECTOR,
            f'div[data-origin="{DOMAIN}"] a.honeybadgerPowered').is_displayed(), (
                "Undo arrow should be displayed")

        # restore control to Badger
        self.open_popup(FIXTURE_URL)
        self.driver.find_element(By.CSS_SELECTOR,
            f'div[data-origin="{DOMAIN}"] a.honeybadgerPowered').click()

        # get back to a valid window handle as the window just got closed
        self.driver.switch_to.window(self.driver.window_handles[0])
        self.open_window()

        # verify the domain is no longer user controlled
        self.load_url(self.options_url)
        self.wait_for_script("return window.OPTIONS_INITIALIZED")
        self.find_el_by_css('a[href="#tab-tracking-domains"]').click()
        action = self.get_domain_slider_state(DOMAIN)
        assert action == "cookieblock", "Domain's action should have been restored"
        assert not self.driver.find_element(By.CSS_SELECTOR,
            f'div[data-origin="{DOMAIN}"] a.honeybadgerPowered').is_displayed(), (
                "Undo arrow should not be displayed")

    def test_disable_enable_buttons(self):
        """Ensure disable/enable buttons change popup state."""

        def get_enable_button():
            return self.driver.find_element(By.ID, "activate_site_btn")

        def get_disable_button():
            return self.driver.find_element(By.ID, "deactivate_site_btn")

        DISPLAYED_ERROR = " should not be displayed on popup"
        NOT_DISPLAYED_ERROR = " should be displayed on popup"
        DUMMY_PAGE_URL = "https://privacybadger-tests.eff.org/html/widget_frame.html"

        self.load_url(DUMMY_PAGE_URL)
        self.open_popup(DUMMY_PAGE_URL)
        get_disable_button().click()

        # get back to a valid window handle as the window just got closed
        self.driver.switch_to.window(self.driver.window_handles[0])
        self.open_popup(DUMMY_PAGE_URL)

        # Check that popup state changed after disabling.
        disable_button = get_disable_button()
        assert not disable_button.is_displayed(), (
            "Disable button" + DISPLAYED_ERROR)
        enable_button = get_enable_button()
        assert enable_button.is_displayed(), (
            "Enable button" + NOT_DISPLAYED_ERROR)

        enable_button.click()

        self.driver.switch_to.window(self.driver.window_handles[0])
        self.open_popup(DUMMY_PAGE_URL)

        # Check that popup state changed after re-enabling.
        disable_button = get_disable_button()
        assert disable_button.is_displayed(), "Disable button" + NOT_DISPLAYED_ERROR
        enable_button = get_enable_button()
        assert not enable_button.is_displayed(), "Enable button" + DISPLAYED_ERROR

    def test_error_button(self):
        """Ensure error button opens report error overlay."""

        DUMMY_PAGE_URL = "https://privacybadger-tests.eff.org/html/widget_frame.html"
        self.load_url(DUMMY_PAGE_URL)
        self.open_popup(DUMMY_PAGE_URL)

        report_input = self.driver.find_element(By.ID, "error-input")
        assert not report_input.is_displayed(), (
            "Error reporting should be closed by default")

        # click the "Report broken site" button
        self.driver.find_element(By.ID, "error").click()
        time.sleep(1)
        assert report_input.is_displayed(), (
            "Error reporting should be open")

        self.driver.find_element(By.ID, "report-close").click()
        time.sleep(1)
        assert not report_input.is_displayed(), (
            "Error reporting should have gotten closed")

        # verify saving and restoring in-progress reports
        ERROR_REPORT_TEXT = "The site isn't loading"
        self.driver.find_element(By.ID, "error").click()
        time.sleep(1)
        report_input.send_keys(ERROR_REPORT_TEXT)
        self.driver.close()
        self.driver.switch_to.window(self.driver.window_handles[0])
        self.open_popup(DUMMY_PAGE_URL)
        time.sleep(1)
        report_input = self.driver.find_element(By.ID, "error-input")
        assert report_input.is_displayed(), "Error reporting should be open"
        assert report_input.get_property('value') == ERROR_REPORT_TEXT, (
            "Previously entered text should be displayed")

    def test_breakage_warnings(self):
        YLIST_DOMAIN = "code.jquery.com"
        FIXTURE_URL = ("https://privacybadger-tests.eff.org/html/"
            "assorted_thirdparties.html")
        self.cookieblock_domain(YLIST_DOMAIN)

        self.load_url(FIXTURE_URL)
        self.open_popup(FIXTURE_URL)

        def get_breakage_icon():
            return self.driver.find_element(By.CSS_SELECTOR,
                f'div.clicker[data-origin="{YLIST_DOMAIN}"] span.breakage-warning')

        # reveal sliders
        self.driver.find_element(By.ID, 'expand-blocked-resources').click()
        self.wait_for_sliders()

        # verify there is no breakage warning
        breakage_icon = get_breakage_icon()
        assert not breakage_icon.is_displayed()

        # manually block the yellowlisted domain
        self.js(f'$("#block-{YLIST_DOMAIN.replace(".", "-")}").click()')

        # verify breakage warning is shown
        breakage_icon = get_breakage_icon()
        assert breakage_icon.is_displayed()

        # verify breakage warning is there when reopened
        self.open_popup(FIXTURE_URL)
        breakage_icon = get_breakage_icon()
        assert breakage_icon.is_displayed()

    def test_slider_hiding(self):
        FIXTURE_URL = ("https://privacybadger-tests.eff.org/html/"
            "assorted_thirdparties.html")
        YLIST_DOMAIN = "code.jquery.com"
        BLOCKED_DOMAIN = "efforg.github.io"

        def assert_hidden(sliders):
            for slider in sliders:
                assert not slider.is_displayed(), (
                    "{slider.get_dom_attribute('data-origin')} is visible but should be hidden")

        def assert_visible(sliders):
            for slider in sliders:
                assert slider.is_displayed(), (
                    f"{slider.get_dom_attribute('data-origin')} is hidden but should be visible")

        self.cookieblock_domain(YLIST_DOMAIN)
        self.block_domain(BLOCKED_DOMAIN)

        self.load_url(FIXTURE_URL)
        self.open_popup(FIXTURE_URL)
        sliders = self.driver.find_elements(By.CSS_SELECTOR, 'div.clicker')

        # verify we have the expected number of sliders
        assert len(sliders) == 2

        # verify sliders are hidden
        assert_hidden(sliders)

        # reveal sliders
        self.driver.find_element(By.ID, 'expand-blocked-resources').click()
        self.wait_for_sliders()

        # verify sliders are visible
        assert_visible(sliders)

        # reopen popup
        self.open_popup(FIXTURE_URL)
        sliders = self.driver.find_elements(By.CSS_SELECTOR, 'div.clicker')

        # verify sliders are visible
        assert_visible(sliders)

        # verify domain is shown second in the list
        assert sliders[1].get_dom_attribute('data-origin') == YLIST_DOMAIN

        # manually block the yellowlisted domain
        self.js(f'$("#block-{YLIST_DOMAIN.replace(".", "-")}").click()')

        # hide sliders
        self.driver.find_element(By.ID, 'collapse-blocked-resources').click()
        self.wait_for_sliders(visible=False)

        # verify sliders are hidden
        assert_hidden(sliders)

        # reopen popup
        self.open_popup(FIXTURE_URL)
        sliders = self.driver.find_elements(By.CSS_SELECTOR, 'div.clicker')

        # verify sliders are visible
        assert_visible(sliders)

        # verify breakage warning slider is at the top
        assert sliders[0].get_dom_attribute('data-origin') == YLIST_DOMAIN

        # restore the user-set slider to default action
        self.driver.find_element(By.CSS_SELECTOR,
            f'div[data-origin="{YLIST_DOMAIN}"] a.honeybadgerPowered').click()

        # get back to a valid window handle as the window just got closed
        self.driver.switch_to.window(self.driver.window_handles[0])

        # reopen popup
        self.open_popup(FIXTURE_URL)
        sliders = self.driver.find_elements(By.CSS_SELECTOR, 'div.clicker')

        # verify sliders are hidden again
        assert_hidden(sliders)

    def test_nothing_blocked_slider_list(self):
        """Verifies display of non-tracking domains."""

        FIXTURE_URL = ("https://privacybadger-tests.eff.org/html/"
            "assorted_thirdparties.html")
        DOMAIN = "efforg.github.io"

        self.load_url(FIXTURE_URL)

        # base case: no sliders should be shown
        self.open_popup(FIXTURE_URL)
        sliders = self.driver.find_elements(By.CSS_SELECTOR, 'div.clicker')
        assert not sliders

        # enable local learning and showing non-tracking domains
        self.load_url(self.options_url)
        self.wait_for_script("return window.OPTIONS_INITIALIZED")
        self.find_el_by_css('a[href="#tab-general-settings"]').click()
        self.find_el_by_css('#local-learning-checkbox').click()
        self.find_el_by_css('#show-nontracking-domains-checkbox').click()

        # non-tracking sliders should now be shown (no collapsing)
        self.open_popup(FIXTURE_URL)
        sliders = self.driver.find_elements(By.CSS_SELECTOR, 'div.clicker')
        assert sliders
        slider = self.driver.find_element(By.CSS_SELECTOR,
            f'div.clicker[data-origin="{DOMAIN}"]')
        assert slider.is_displayed()
        assert self.get_domain_slider_state(DOMAIN) == "allow"

        # now block the domain and verify blocked slider is hidden (list is collapsed)
        self.close_window_with_url(FIXTURE_URL)
        self.block_domain(DOMAIN)
        self.load_url(FIXTURE_URL)
        self.open_popup(FIXTURE_URL)
        slider = self.driver.find_element(By.CSS_SELECTOR,
            f'div.clicker[data-origin="{DOMAIN}"]')
        assert not slider.is_displayed()


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python

import unittest

import pbtest

from functools import partial


class SupercookieTest(pbtest.PBSeleniumTest):
    """Make sure we detect potential supercookies. """

    def get_snitch_map_for(self, domain):
        self.open_window() # don't replace the test page to allow for retrying
        return self.get_badger_storage('snitch_map').get(domain)

    def setUp(self):
        # enable local learning
        self.load_url(self.options_url)
        self.wait_for_script("return window.OPTIONS_INITIALIZED")
        self.find_el_by_css('a[href="#tab-general-settings"]').click()
        self.find_el_by_css('#local-learning-checkbox').click()

    def test_should_detect_ls_of_third_party_frame(self):
        FIRST_PARTY_BASE = "eff.org"
        THIRD_PARTY_BASE = "efforg.github.io"

        assert not self.get_snitch_map_for(THIRD_PARTY_BASE)

        self.load_url((
            f"https://privacybadger-tests.{FIRST_PARTY_BASE}/html/"
            "localstorage.html"
        ))

        # TODO FIXME We get some intermittent failures for this test.
        # It seems we sometimes miss the setting of localStorage items
        # because the script runs after we already checked what's in localStorage.
        # We can work around this race condition by reloading the page.
        self.driver.refresh()

        snitch_map = pbtest.retry_until(
            partial(self.get_snitch_map_for, THIRD_PARTY_BASE),
            times=3)
        assert snitch_map == [FIRST_PARTY_BASE]

    def test_should_not_detect_low_entropy_ls_of_third_party_frame(self):
        FIRST_PARTY_BASE = "eff.org"
        THIRD_PARTY_BASE = "efforg.github.io"
        assert not self.get_snitch_map_for(THIRD_PARTY_BASE)
        self.load_url((
            f"https://privacybadger-tests.{FIRST_PARTY_BASE}/html/"
            "localstorage_low_entropy.html"
        ))
        self.driver.refresh() # TODO workaround
        assert not self.get_snitch_map_for(THIRD_PARTY_BASE)

    def test_should_not_detect_first_party_ls(self):
        BASE_DOMAIN = "efforg.github.io"
        self.load_url((
            f"https://{BASE_DOMAIN}/privacybadger-test-fixtures/html/"
            "localstorage/set_ls.html"
        ))
        self.driver.refresh() # TODO workaround
        assert not self.get_snitch_map_for(BASE_DOMAIN)

    def test_should_not_detect_ls_of_third_party_script(self):
        FIRST_PARTY_BASE = "eff.org"
        THIRD_PARTY_BASE = "efforg.github.io"

        # a third-party script included by the top page (not a 3rd party frame)
        self.load_url((
            f"https://privacybadger-tests.{FIRST_PARTY_BASE}/html/"
            "localstorage_from_third_party_script.html"
        ))

        self.driver.refresh() # TODO workaround

        assert not self.get_snitch_map_for(FIRST_PARTY_BASE)
        assert not self.get_snitch_map_for(THIRD_PARTY_BASE)

    def test_localstorage_learning(self):
        """Verifies that we learn to block a third-party domain if we see
        non-trivial localStorage data from that third-party on three sites."""

        SITE1_URL = "https://ddrybktjfxh4.cloudfront.net/localstorage.html"
        SITE2_URL = "https://d3syxqe9po5ji0.cloudfront.net/localstorage.html"
        SITE3_URL = "https://d3b37ucnz1m2l2.cloudfront.net/localstorage.html"

        THIRD_PARTY = "efforg.github.io"

        # remove pre-trained domains
        self.clear_tracker_data()

        def get_sliders(url, category):
            self.open_popup(url)
            sliders = self.get_tracker_state()
            return sliders[category]

        # load the first site
        self.load_url(SITE1_URL)
        # TODO FIXME remove workaround once we fix race conditions in contentscripts/supercookie.js
        self.driver.refresh()
        sliders = pbtest.retry_until(
            partial(get_sliders, SITE1_URL, 'notYetBlocked'), times=3)
        assert THIRD_PARTY in sliders

        # go to second site
        self.load_url(SITE2_URL)
        # TODO workaround
        self.driver.refresh()
        sliders = pbtest.retry_until(
            partial(get_sliders, SITE2_URL, 'notYetBlocked'), times=3)
        assert THIRD_PARTY in sliders

        # go to third site
        self.load_url(SITE3_URL)
        # TODO workaround
        self.driver.refresh()
        sliders = pbtest.retry_until(
            partial(get_sliders, SITE3_URL, 'blocked'), times=3)
        assert THIRD_PARTY in sliders, "third party should now be reported as blocked"


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python

import unittest

import pbtest

from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import WebDriverWait


class ServiceWorkersTest(pbtest.PBSeleniumTest):
    """Verifies interaction with sites that use Service Worker caches"""

    FIXTURE_URL = (
        "https://efforg.github.io/privacybadger-test-fixtures/html/"
        "service_workers.html"
    )
    FIXTURE_HOST = "efforg.github.io"
    THIRD_PARTY_HOST = "privacybadger-tests.eff.org"

    def get_first_tab_data(self):
        """Returns the top-level document's URL/host for the first tab."""
        tab_data = self.driver.execute_async_script(
            "let done = arguments[arguments.length - 1];"
            "chrome.runtime.sendMessage({ type: 'getTabData' }, tabData => {"
            "  let min_tab_id = Math.min(...Object.keys(tabData));"
            "  done(tabData[min_tab_id].frames[0]);"
            "});")
        return tab_data

    def init_sw_page(self):
        # visit the Service Worker page to activate the worker
        self.load_url(self.FIXTURE_URL)

        # Service Workers are off by default in Firefox 60 ESR
        if not self.js("return 'serviceWorker' in navigator;"):
            self.skipTest("Service Workers are disabled")

        # wait for the worker to initialize its cache
        self.wait_for_script("return window.WORKER_READY;")

        # visit a different site (doesn't matter what it is,
        # just needs to be an http site with a different domain)
        self.load_url("https://dnt-test.trackersimulator.org/")

    def test_returning_to_sw_cached_page(self):
        self.init_sw_page()

        # return to the SW page
        self.driver.back()

        # open new window (to avoid clearing badger.tabData) and verify results
        self.open_window()
        self.load_url(self.options_url)
        tab_data = self.get_first_tab_data()
        assert tab_data['host'] == self.FIXTURE_HOST, (
            "Unexpected first-tab hostname in tabData")

    def test_redirect_to_sw_cached_page(self):
        self.init_sw_page()

        # visit a page that 302-redirects back to our fixture
        self.load_url("https://www.eff.org/privacy-badger-service-workers-navigation-test-redirect")

        # wait for URL to change
        WebDriverWait(self.driver, 10).until(EC.url_to_be(self.FIXTURE_URL))

        # open new window (to avoid clearing badger.tabData) and verify results
        self.open_window()
        self.load_url(self.options_url)
        tab_data = self.get_first_tab_data()
        assert tab_data['url'] == self.FIXTURE_URL, (
            "Unexpected first-tab URL in tabData")

    def test_blocking_sw_initiated_requests(self):
        self.clear_tracker_data()
        self.block_domain(self.THIRD_PARTY_HOST)

        self.init_sw_page()

        self.load_url(self.FIXTURE_URL)

        # check what happened
        selector = '#third-party-load-result'
        self.wait_for_script(
            "return document.querySelector(arguments[0]).textContent",
            selector)
        assert self.find_el_by_css(selector).text == "error", (
            "third-party should have gotten blocked")

        # also check the popup
        self.open_popup(self.FIXTURE_URL)
        sliders = self.get_tracker_state()
        assert self.THIRD_PARTY_HOST in sliders['blocked'], (
            "third-party should be reported as blocked")

    def test_disabling_on_site(self):
        self.clear_tracker_data()
        self.block_domain(self.THIRD_PARTY_HOST)

        self.init_sw_page()

        self.disable_badger_on_site(self.FIXTURE_HOST)

        self.load_url(self.FIXTURE_URL)

        # check what happened
        selector = '#third-party-load-result'
        self.wait_for_script(
            "return document.querySelector(arguments[0]).textContent",
            selector)
        assert self.find_el_by_css(selector).text == "success", (
            "third-party should have been allowed to load")


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3

import json
import random
import re
import sys
import time
import urllib.request

from collections import OrderedDict

from tldextract import TLDExtract


DNT_LIST_URL = 'https://www.eff.org/files/effdntlist.txt'


def get_dnt_domains():
    dnt_domains = []

    with urllib.request.urlopen(DNT_LIST_URL) as response:
        domain_regex = r'^@@\|\|(.*)\^\$third-party$'
        for line in response:
            match = re.match(domain_regex, line.decode('utf-8').strip())
            if match:
                dnt_domains.append(match.groups()[0])

    return dnt_domains

def get_next_update_time():
    epoch_now = int(time.time()) * 1000
    one_day = 1000 * 60 * 60 * 24
    return random.randint(epoch_now + one_day, epoch_now + one_day * 7)

def apply_dnt_domains(data, dnt_domains):
    tld_extract = TLDExtract(cache_dir=False)

    for domain in dnt_domains:
        parsed_tld = tld_extract(domain)
        base = parsed_tld.domain + '.' + parsed_tld.suffix

        # only apply DNT scan results for domains
        # that are already present in seed data
        # and are blocked
        if base not in data['action_map'] or data['action_map'][base]['heuristicAction'] != "block":
            continue

        if domain in data['action_map']:
            if "dnt" not in data['action_map'][domain] or not data['action_map'][domain]['dnt']:
                print("Marking %s as DNT compliant ..." % domain)
                data['action_map'][domain]['dnt'] = True
                data['action_map'][domain]['nextUpdateTime'] = get_next_update_time()
        else:
            print("Adding %s as DNT compliant ..." % domain)
            data['action_map'][domain] = {
                "dnt": True,
                "heuristicAction": "",
                "nextUpdateTime": get_next_update_time(),
            }

    return data


if __name__ == "__main__":
    if len(sys.argv) != 2:
        print("Usage: %s BADGER_SEED_DATA.json" % sys.argv[0])
        sys.exit(1)

    # get DNT scan results
    dnt_domains = get_dnt_domains()
    if not dnt_domains:
        print("No DNT list domains loaded!")
        sys.exit(1)

    with open(sys.argv[1], 'r+') as seed_file:
        # read in seed data, preserving ordering
        seed_data = json.load(seed_file, object_pairs_hook=OrderedDict)

        # apply DNT scan results to seed data
        seed_data = apply_dnt_domains(seed_data, dnt_domains)

        # write the data back out
        seed_file.seek(0)
        # this should match how data gets written out by Badger Sett
        json.dump(seed_data, seed_file, indent=2, sort_keys=True, separators=(',', ': '))



#!/usr/bin/env python3

import json

from collections import OrderedDict
from glob import glob

SOURCE_LOCALE = 'src/_locales/en_US/messages.json'


def fix_locale(locale, placeholder_keys):
    # read in locale, preserving existing ordering
    with open(locale, 'r', encoding="utf-8") as f:
        data = json.load(f, object_pairs_hook=OrderedDict)

    # restore missing placeholders
    for key in placeholder_keys:
        if key in data and "placeholders" not in data[key]:
            data[key]["placeholders"] = source_data[key]["placeholders"]

    with open(locale, 'w', encoding="utf-8") as f:
        json.dump(data, f, ensure_ascii=False, indent=4)


if __name__ == '__main__':
    with open(SOURCE_LOCALE, 'r', encoding="utf-8") as f:
        source_data = json.load(f, object_pairs_hook=OrderedDict)

    # get keys of locale messages with placeholders
    placeholder_keys = []
    for key in source_data:
        if "placeholders" in source_data[key]:
            placeholder_keys.append(key)

    # fix all locales
    for locale in glob('src/_locales/*/*.json'):
        if locale == SOURCE_LOCALE:
            continue

        fix_locale(locale, placeholder_keys)



#!/usr/bin/env python

import unittest

import pytest

import pbtest

from functools import partial


class NavigationTest(pbtest.PBSeleniumTest):
    """Verifies navigation-related fixes and workarounds."""

    def get_trackers(self, url):
        trackers = self.driver.execute_async_script(
            "let done = arguments[arguments.length - 1];"
            "chrome.tabs.query({ url: arguments[0] }, function (tabs) {"
            "  if (!tabs[0]) {"
            "    done(null);"
            "  }"
            "  chrome.runtime.sendMessage({ type: 'getTabData' }, tabData => {"
            "    done(tabData[tabs[0].id].trackers);"
            "  });"
            "});", url)
        return trackers

    @pytest.mark.xfail(pbtest.shim.browser_type == "edge",
                       reason="https://github.com/microsoft/MicrosoftEdge-Extensions/issues/228")
    def test_beacon_attribution(self):
        FIXTURE_URL = (
            "https://efforg.github.io/privacybadger-test-fixtures/html/"
            "beacon.html"
        )
        OTHER_SITE_URL = "https://dnt-test.trackersimulator.org/"
        THIRD_PARTY_HOST = "privacybadger-tests.eff.org"

        self.clear_tracker_data()
        # enable local learning
        self.wait_for_script("return window.OPTIONS_INITIALIZED")
        self.find_el_by_css('a[href="#tab-general-settings"]').click()
        self.find_el_by_css('#local-learning-checkbox').click()

        self.load_url(FIXTURE_URL)
        # open new window (to avoid clearing badger.tabData) and verify results
        self.open_window()
        self.load_url(self.options_url)
        assert self.get_trackers(FIXTURE_URL) == {}, "beacon should not have fired yet"

        # verify beacon domain is listed after page refresh
        self.driver.switch_to.window(self.driver.window_handles[-2])
        self.driver.refresh()
        self.driver.switch_to.window(self.driver.window_handles[-1]) 
        domains = pbtest.retry_until(partial(self.get_trackers, FIXTURE_URL),
                                     tester=lambda x: x == {THIRD_PARTY_HOST: "allow"},
                                     times=3)
        assert domains == {THIRD_PARTY_HOST: "allow"}, "beacon should have fired"

        # visit a different site (doesn't matter what it is,
        # just needs to be an http site with a different domain)
        self.driver.switch_to.window(self.driver.window_handles[-2])
        self.load_url(OTHER_SITE_URL)
        # verify no domains are listed
        self.driver.switch_to.window(self.driver.window_handles[-1])
        assert self.get_trackers(OTHER_SITE_URL) == {}, "beacon got wrongly attributed to an unrelated site"


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python

import unittest

import pbtest


class CookieTest(pbtest.PBSeleniumTest):
    """Basic test to make sure the PB doesn't mess up with the cookies."""

    def assert_pass_opera_cookie_test(self, url, test_name):
        self.load_url(url)
        assert self.txt_by_css("#result") == "PASS", (
            f"Cookie test failed: {test_name}")

    def test_should_pass_std_cookie_test(self):
        self.assert_pass_opera_cookie_test((
            "https://efforg.github.io/privacybadger-test-fixtures/html/"
            "first_party_cookie.html"
        ), "Set 1st party cookie")

    def test_cookie_tracker_detection(self):
        """Tests basic cookie tracking. The tracking site has no DNT file,
        and gets blocked by PB.

        Visits three sites, all of which have an iframe that points to a fourth site
        that reads and writes a cookie. The third party cookie will be picked up by
        PB after each of the site loads, but no action will be taken. Then the first
        site will be reloaded, and the UI will show the third party domain as blocked."""

        SITE1_URL = "https://ddrybktjfxh4.cloudfront.net/"
        SITE2_URL = "https://d3syxqe9po5ji0.cloudfront.net/"
        SITE3_URL = "https://d3b37ucnz1m2l2.cloudfront.net/"

        THIRD_PARTY_DOMAIN = "efforg.github.io"

        # remove pre-trained domains
        self.clear_tracker_data()

        # enable local learning
        self.wait_for_script("return window.OPTIONS_INITIALIZED")
        self.find_el_by_css('a[href="#tab-general-settings"]').click()
        self.find_el_by_css('#local-learning-checkbox').click()

        # load the first site with the third party code that reads and writes a cookie
        self.load_url(SITE1_URL)
        self.open_popup(SITE1_URL)
        # TODO it takes another visit (or a page reload)
        # TODO to show the domain as not-yet-blocked-but-tracking?
        #sliders = self.get_tracker_state()
        #assert THIRD_PARTY_DOMAIN in sliders['notYetBlocked']
        self.close_window_with_url(SITE1_URL)

        # go to second site
        self.load_url(SITE2_URL)
        self.open_popup(SITE2_URL)
        sliders = self.get_tracker_state()
        assert THIRD_PARTY_DOMAIN in sliders['notYetBlocked']
        self.close_window_with_url(SITE2_URL)

        # go to third site
        self.load_url(SITE3_URL)
        self.open_popup(SITE3_URL)
        sliders = self.get_tracker_state()
        assert THIRD_PARTY_DOMAIN in sliders['notYetBlocked']
        self.close_window_with_url(SITE3_URL)

        # revisiting the first site should cause
        # the third-party domain to be blocked
        self.load_url(SITE1_URL)
        self.open_popup(SITE1_URL)
        sliders = self.get_tracker_state()
        assert THIRD_PARTY_DOMAIN in sliders['blocked']


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python

import unittest

import pytest

import pbtest

from selenium.common.exceptions import (
    NoSuchElementException,
    TimeoutException,
)
from selenium.webdriver.common.by import By
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.select import Select
from selenium.webdriver.support.ui import WebDriverWait


class OptionsTest(pbtest.PBSeleniumTest):
    """Make sure the options page works correctly."""

    def assert_domain_toggle_state(self, domain, action, failure_msg):
        clicker = self.find_el_by_css(f'div[data-origin="{domain}"]')
        assert clicker.get_dom_attribute("class") == "clicker userset", failure_msg

        switches_div = clicker.find_element(By.CSS_SELECTOR, ".switch-container")
        assert switches_div.get_dom_attribute("class") == f"switch-container {action}", failure_msg

    def find_domain_toggle_by_xpath(self, domain):
        toggles_div = self.driver.find_element(By.ID, "blockedResourcesInner")
        return toggles_div.find_element(By.XPATH, (
            './/div[@data-origin="{domain}"]'
            # test that "origin" is one of the classes on the element:
            # https://stackoverflow.com/a/1390680
            '//div[contains(concat(" ", normalize-space(@class), " "), " origin ")]'
            '//span[text()="{domain}"]'
        ).format(domain=domain))

    def select_domain_list_tab(self):
        self.find_el_by_css('a[href="#tab-tracking-domains"]').click()

    def select_manage_data_tab(self):
        self.find_el_by_css('a[href="#tab-manage-data"]').click()

    def load_options_page(self):
        self.load_url(self.options_url + '?all')
        self.wait_for_script("return window.OPTIONS_INITIALIZED")

    def test_reloading_should_reapply_filters(self):
        FILTERVAL = "user"

        self.load_options_page()
        self.select_domain_list_tab()

        # change a domain list filter
        Select(self.find_el_by_css('#tracking-domains-type-filter')).select_by_value(FILTERVAL)

        # reload page and assert filters are set
        self.driver.refresh()
        sel = Select(self.find_el_by_css('#tracking-domains-type-filter'))
        assert sel.first_selected_option.get_property('value') == FILTERVAL

        # open options page in a new window and assert filters are not set
        self.open_window()
        self.load_options_page()
        self.select_domain_list_tab()
        sel = Select(self.find_el_by_css('#tracking-domains-type-filter'))
        assert not sel.first_selected_option.get_property('value')

    @pytest.mark.flaky(reruns=5, condition=pbtest.shim.browser_type == "edge")
    def test_adding_domain(self):
        """Ensure domain and tracker count are displayed."""
        self.clear_tracker_data()

        self.add_domain("pbtest.org", "block")

        self.load_options_page()
        self.select_domain_list_tab()

        error_message = "The 'multiple tracker' message should be displayed after adding a domain"
        assert self.driver.find_element(By.ID, "options_domain_list_trackers").is_displayed(), error_message
        assert not self.driver.find_element(By.ID, "options_domain_list_no_trackers").is_displayed(), error_message

        try:
            self.find_domain_toggle_by_xpath("pbtest.org")
        except NoSuchElementException:
            self.fail("Tracking domain is not displayed")

    @pytest.mark.flaky(reruns=5, condition=pbtest.shim.browser_type == "edge")
    def test_removing_domain(self):
        """Ensure domain is removed properly."""
        self.clear_tracker_data()
        self.add_domain("pbtest.org", "block")

        self.load_options_page()
        self.select_domain_list_tab()

        domains = self.driver.find_elements(By.CSS_SELECTOR, 'div.clicker')
        assert len(domains) == 1, "Should see exactly one domain in the list"

        # remove displayed domain
        remove_domain_element = self.find_el_by_xpath(
            './/div[@data-origin="pbtest.org"]'
            '//a[@class="removeDomain"]')
        remove_domain_element.click()

        # Make sure the alert is present. Otherwise we get intermittent errors.
        WebDriverWait(self.driver, 3).until(EC.alert_is_present())
        self.driver.switch_to.alert.accept()

        # verify that only the 'no trackers' message is displayed
        try:
            WebDriverWait(self.driver, 5).until(
                EC.visibility_of_element_located((By.ID, "options_domain_list_no_trackers")))
        except TimeoutException:
            self.fail("There should be a 'no trackers' message after deleting domain")

        error_message = "Only the 'no trackers' message should be displayed"
        assert not self.driver.find_element(By.ID, "options_domain_list_trackers").is_displayed(), error_message

        # verify that no domains are displayed
        try:
            domains = self.driver.find_elements(By.CSS_SELECTOR, 'div.clicker')
        except NoSuchElementException:
            domains = []
        assert len(domains) == 0, "No domains should be displayed after removal"

    def test_reset_data(self):
        self.load_options_page()
        self.select_domain_list_tab()

        # make sure the default tracker list includes many trackers
        error_message = "By default, the tracker list should contain many trackers"
        assert self.driver.find_element(By.ID, "options_domain_list_trackers").is_displayed(), error_message
        assert not self.driver.find_element(By.ID, "options_domain_list_no_trackers").is_displayed(), error_message

        # get the number of trackers in the seed data
        default_summary_text = self.driver.find_element(By.ID, "options_domain_list_trackers").text

        # Click on the "remove all data" button to empty the tracker lists, and
        # click "OK" in the popup that ensues
        self.select_manage_data_tab()
        with self.wait_for_reload():
            self.driver.find_element(By.ID, 'removeAllData').click()
            self.driver.switch_to.alert.accept()

        # now make sure the tracker list is empty
        self.select_domain_list_tab()
        error_message = "No trackers should be displayed after removing all data"
        assert not self.driver.find_element(By.ID, "options_domain_list_trackers").is_displayed(), error_message
        assert self.driver.find_element(By.ID, "options_domain_list_no_trackers").is_displayed(), error_message

        # add new blocked domains
        self.add_domain("pbtest.org", "block")
        self.add_domain("pbtest1.org", "block")

        # reload the options page
        self.load_options_page()
        self.select_domain_list_tab()

        # make sure only two trackers are displayed now
        actual_text = self.driver.find_element(By.ID, "options_domain_list_trackers").text
        expected_text = self.js("return chrome.i18n.getMessage("
            "'options_domain_list_trackers', [2, '']);").replace("</a>", "")
        assert actual_text == expected_text

        # click the "reset data" button to restore seed data and get rid of the
        # domains we learned
        self.select_manage_data_tab()
        with self.wait_for_reload():
            self.driver.find_element(By.ID, 'resetData').click()
            self.driver.switch_to.alert.accept()

        # make sure the same number of trackers are displayed as by default
        self.select_domain_list_tab()
        error_message = "After resetting data, tracker count should return to default"
        assert self.driver.find_element(By.ID, "options_domain_list_trackers").text == default_summary_text, error_message

    def tracking_user_overwrite(self, original_action, overwrite_action):
        """Ensure preferences are persisted when a user overwrites
        PB's default behaviour for a domain."""

        DOMAIN = "pbtest.org"

        self.clear_tracker_data()
        self.add_domain(DOMAIN, original_action)

        self.load_options_page()
        self.wait_for_script("return window.OPTIONS_INITIALIZED")
        self.find_el_by_css('a[href="#tab-general-settings"]').click()
        # enable learning to reveal the show-not-yet-blocked checkbox
        self.find_el_by_css('#local-learning-checkbox').click()
        self.select_domain_list_tab()
        self.find_el_by_css('#tracking-domains-show-not-yet-blocked').click()
        # wait for toggles to finish rendering
        self.wait_for_script("return window.SLIDERS_DONE")

        # Change user preferences
        domain_id = DOMAIN.replace(".", "-")
        self.wait_for_script(f"return $('#{overwrite_action}-{domain_id}')[0];")
        self.js(f"$('#{overwrite_action}-{domain_id}').trigger('click');")

        # Re-open the tab
        self.load_options_page()
        self.select_domain_list_tab()
        self.find_el_by_css('#tracking-domains-show-not-yet-blocked').click()
        # wait for toggles to finish rendering
        self.wait_for_script("return window.SLIDERS_DONE")

        # Check the user preferences for the domains are still displayed
        failure_msg = (
            f"Domain should be displayed as {overwrite_action} "
            f"after user overwrite of PB's decision to {original_action}")
        self.assert_domain_toggle_state(DOMAIN, overwrite_action, failure_msg)

    @pytest.mark.flaky(reruns=5, condition=pbtest.shim.browser_type in ("chrome", "edge"))
    def test_tracking_user_overwrite_allowed_block(self):
        self.tracking_user_overwrite('allow', 'block')

    @pytest.mark.flaky(reruns=5, condition=pbtest.shim.browser_type in ("chrome", "edge"))
    def test_tracking_user_overwrite_allowed_cookieblock(self):
        self.tracking_user_overwrite('allow', 'cookieblock')

    @pytest.mark.flaky(reruns=5, condition=pbtest.shim.browser_type in ("chrome", "edge"))
    def test_tracking_user_overwrite_cookieblocked_allow(self):
        self.tracking_user_overwrite('cookieblock', 'allow')

    @pytest.mark.flaky(reruns=5, condition=pbtest.shim.browser_type in ("chrome", "edge"))
    def test_tracking_user_overwrite_cookieblocked_block(self):
        self.tracking_user_overwrite('cookieblock', 'block')

    @pytest.mark.flaky(reruns=5, condition=pbtest.shim.browser_type in ("chrome", "edge"))
    def test_tracking_user_overwrite_blocked_allow(self):
        self.tracking_user_overwrite('block', 'allow')

    @pytest.mark.flaky(reruns=5, condition=pbtest.shim.browser_type in ("chrome", "edge"))
    def test_tracking_user_overwrite_blocked_cookieblock(self):
        self.tracking_user_overwrite('block', 'cookieblock')

    # early-warning check for the open_in_tab attribute of options_ui
    # https://github.com/EFForg/privacybadger/pull/1775#pullrequestreview-76940251
    def test_options_ui_open_in_tab(self):
        # open options page manually
        self.open_window()
        self.load_options_page()

        # open the new user intro page
        self.open_window()
        self.load_url(self.first_run_url)

        # save open windows
        handles_before = set(self.driver.window_handles)

        # open options page using dedicated chrome API
        self.js("chrome.runtime.openOptionsPage();")

        # if we switched to the previously manually opened options page, all good
        # if we haven't, this must mean options_ui's open_in_tab no longer works
        new_handles = set(self.driver.window_handles).difference(handles_before)
        num_newly_opened_windows = len(new_handles)

        if num_newly_opened_windows:
            self.driver.switch_to.window(new_handles.pop())

        assert num_newly_opened_windows == 0, (
            "Expected to switch to existing options page, "
            f"opened a new page ({self.driver.title}) instead: "
            f"{self.driver.current_url}")


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python

import time
import unittest

import pbtest


class ContentFilteringTest(pbtest.PBSeleniumTest):
    """Content filtering tests."""

    # for blocking tests
    FIXTURE_DOMAIN = "efforg.github.io"
    FIXTURE_PARENT_DOMAIN = "github.io"
    FIXTURE_URL = (
        f"https://{FIXTURE_DOMAIN}/privacybadger-test-fixtures/html/"
        "3p_script_with_load_status.html"
    )
    THIRD_PARTY_DOMAIN = "privacybadger-tests.eff.org"
    THIRD_PARTY_SUBDOMAIN = "child1." + THIRD_PARTY_DOMAIN
    SELECTOR = "#third-party-load-result"

    # for cookie tests
    COOKIE_FIXTURE_URL = (
        "https://efforg.github.io/privacybadger-test-fixtures/html/"
        "recording_nontracking_domains.html"
    )
    COOKIE_DOMAIN = "dnt-request-cookies-test.trackersimulator.org"

    def setUp(self):
        self.clear_tracker_data()

    def assert_block(self):
        self.wait_for_any_text(self.SELECTOR)
        assert self.find_el_by_css(self.SELECTOR).text == "error", (
            "3rd-party should've gotten blocked")

    def assert_load(self):
        self.wait_for_any_text(self.SELECTOR)
        assert self.find_el_by_css(self.SELECTOR).text == "success", (
            "3rd-party should've loaded successfully")

    def test_blocking_fixture_loads_when_unblocked(self):
        self.load_url(self.FIXTURE_URL)
        self.assert_load()

    def test_blocking(self):
        self.block_domain(self.THIRD_PARTY_DOMAIN)

        self.load_url(self.FIXTURE_URL)
        self.assert_block()

    def test_cookieblocking_stops_saving(self):
        self.load_url(f"https://{self.COOKIE_DOMAIN}/")
        assert not self.driver.get_cookies(), (
            "Visiting the domain directly does not set a cookie")

        self.load_url(self.COOKIE_FIXTURE_URL)
        self.load_url(f"https://{self.COOKIE_DOMAIN}/")
        assert len(self.driver.get_cookies()) == 1, (
            "Cookie fixture should have set a cookie")

        self.driver.delete_all_cookies()
        self.cookieblock_domain(self.COOKIE_DOMAIN)

        self.load_url(self.COOKIE_FIXTURE_URL)
        self.load_url(f"https://{self.COOKIE_DOMAIN}/")
        assert not self.driver.get_cookies(), (
            "Cookie fixture should have been blocked from setting a cookie")

    def test_cookieblocking_stops_sending(self):
        self.load_url(self.COOKIE_FIXTURE_URL)
        self.wait_for_and_switch_to_frame("iframe[src]", timeout=1)
        self.wait_for_any_text('body')
        assert self.find_el_by_css('body').text == "cookies=0", (
            "No cookies should've been sent to start with")

        self.load_url(self.COOKIE_FIXTURE_URL)
        self.wait_for_and_switch_to_frame("iframe[src]", timeout=1)
        self.wait_for_any_text('body')
        assert self.find_el_by_css('body').text == "cookies=1", (
            "We should have sent a cookie at this point")

        self.cookieblock_domain(self.COOKIE_DOMAIN)

        self.load_url(self.COOKIE_FIXTURE_URL)
        self.wait_for_and_switch_to_frame("iframe[src]", timeout=1)
        self.wait_for_any_text('body')
        assert self.find_el_by_css('body').text == "cookies=0", (
            "No cookies should have been sent by the cookieblocked domain")

    def test_cookieblocking_subdomain_of_blocked_domain(self):
        self.block_domain(self.THIRD_PARTY_DOMAIN)
        self.cookieblock_domain(self.THIRD_PARTY_SUBDOMAIN)
        self.load_url(self.FIXTURE_URL + '?alt3p')
        self.assert_load()

    def test_cookieblocking_base_overwrites_subdomain_block(self):
        self.block_domain(self.THIRD_PARTY_SUBDOMAIN)
        self.load_url(self.FIXTURE_URL + '?alt3p')
        self.assert_block()

        self.cookieblock_domain(self.THIRD_PARTY_DOMAIN)
        self.load_url(self.FIXTURE_URL + '?alt3p')
        self.assert_load()

    def test_blocking_fp_script_from_cookieblocked_cdn(self):
        fp2_fixture = "https://efforg.github.io/privacybadger-test-fixtures/html/fingerprinting.html"

        self.cookieblock_domain("cdn.jsdelivr.net")

        # no canvas script data to start with
        assert self.get_badger_storage('fp_scripts') == {}

        # enable local learning
        self.wait_for_script("return window.OPTIONS_INITIALIZED")
        self.find_el_by_css('a[href="#tab-general-settings"]').click()
        self.find_el_by_css('#local-learning-checkbox').click()

        self.load_url(fp2_fixture)
        assert self.js("return typeof window.Fingerprint2") == "function"

        # keep the window open for recording to complete
        self.open_window()

        # wait for canvas script path to be recorded
        if self.get_badger_storage('fp_scripts') == {}:
            time.sleep(1)

        self.load_url(fp2_fixture)
        self.driver.refresh()

        if self.js("return typeof window.Fingerprint2") != "undefined":
            # clear webRequest caches
            self.load_url(self.options_url)
            self.wait_for_script(
                "(async function (done) {"
                "  await chrome.webRequest.handlerBehaviorChanged();"
                "  done(true);"
                "}(arguments[arguments.length - 1]));", execute_async=True)
            self.load_url(fp2_fixture)
            self.driver.refresh()

        assert self.js("return typeof window.Fingerprint2") == "undefined"

    def test_surrogating_fp_script_from_cookieblocked_cdn(self):
        """Since we have a surrogate script for FingerprintJS (v3) served
        from cdn.jsdelivr.net, we test surrogation rather than blocking."""

        def get_visitor_id():
            return self.driver.execute_async_script(
                "let done = arguments[arguments.length - 1];"
                "FingerprintJS.load().then(fp => {"
                "  fp.get().then(res => {"
                "    done(res.visitorId);"
                "  });"
                "});")

        self.cookieblock_domain("cdn.jsdelivr.net")

        # first get a visitor ID from FingerprintJS, before enabling learning
        self.load_url(self.FIXTURE_URL + '?fingerprintjs')
        self.assert_load()
        visitor_id = get_visitor_id()

        # enable local learning
        self.load_url(self.options_url)
        self.wait_for_script("return window.OPTIONS_INITIALIZED")
        self.find_el_by_css('a[href="#tab-general-settings"]').click()
        self.find_el_by_css('#local-learning-checkbox').click()

        # assert the original script still loads
        self.load_url(self.FIXTURE_URL + '?fingerprintjs')
        self.assert_load()
        assert get_visitor_id() == visitor_id, (
            "Visitor ID should be consistent between page loads")

        # keep the window open for recording to complete
        self.open_window()

        # now assert the surrogate script loads
        self.load_url(self.FIXTURE_URL + '?fingerprintjs')
        self.assert_load()

        if get_visitor_id() == visitor_id:
            # clear webRequest caches
            self.load_url(self.options_url)
            self.wait_for_script(
                "(async function (done) {"
                "  await chrome.webRequest.handlerBehaviorChanged();"
                "  done(true);"
                "}(arguments[arguments.length - 1]));", execute_async=True)
            self.load_url(self.FIXTURE_URL + '?fingerprintjs')
            self.driver.refresh()
            self.assert_load()

        assert get_visitor_id() != visitor_id, (
            "Visitor ID should change between page loads")
        self.driver.refresh()
        self.assert_load()
        assert get_visitor_id() != visitor_id, (
            "Visitor ID should change between page loads")

    def test_userblock(self):
        self.set_user_action(self.THIRD_PARTY_DOMAIN, "block")

        self.load_url(self.FIXTURE_URL)
        self.assert_block()

    def test_userblock_overwrites_cookieblock(self):
        self.cookieblock_domain(self.THIRD_PARTY_DOMAIN)
        self.set_user_action(self.THIRD_PARTY_DOMAIN, "block")
        self.load_url(self.FIXTURE_URL)
        self.assert_block()

    def test_userblock_overwrites_dnt(self):
        self.set_dnt(self.THIRD_PARTY_DOMAIN)
        self.set_user_action(self.THIRD_PARTY_DOMAIN, "block")
        self.load_url(self.FIXTURE_URL)
        self.assert_block()

    def test_usercookieblock_stops_saving(self):
        self.set_user_action(self.COOKIE_DOMAIN, "cookieblock")

        self.load_url(self.COOKIE_FIXTURE_URL)
        self.load_url(f"https://{self.COOKIE_DOMAIN}/")
        assert not self.driver.get_cookies(), (
            "Cookie fixture should have been blocked from setting a cookie")

    def test_usercookieblock_overwrites_block(self):
        self.block_domain(self.THIRD_PARTY_DOMAIN)
        self.load_url(self.FIXTURE_URL)
        self.assert_block()

        self.set_user_action(self.THIRD_PARTY_DOMAIN, "cookieblock")
        self.load_url(self.FIXTURE_URL)
        self.assert_load()

    def test_userallow(self):
        self.block_domain(self.THIRD_PARTY_DOMAIN)
        self.set_user_action(self.THIRD_PARTY_DOMAIN, "allow")

        self.load_url(self.FIXTURE_URL)
        self.assert_load()

    def test_userallow_overwrites_block(self):
        self.block_domain(self.THIRD_PARTY_DOMAIN)
        self.load_url(self.FIXTURE_URL)
        self.assert_block()

        self.set_user_action(self.THIRD_PARTY_DOMAIN, "allow")
        self.load_url(self.FIXTURE_URL)
        self.assert_load()

    def test_reverting_control(self):
        self.block_domain(self.THIRD_PARTY_DOMAIN)
        self.set_user_action(self.THIRD_PARTY_DOMAIN, "allow")
        self.load_url(self.FIXTURE_URL)
        self.assert_load()

        # click the undo arrow
        self.open_popup(self.FIXTURE_URL)
        self.find_el_by_css(
            f'div[data-origin="{self.THIRD_PARTY_DOMAIN}"] a.honeybadgerPowered').click()
        self.driver.switch_to.window(self.driver.window_handles[0])
        self.open_window()

        self.load_url(self.FIXTURE_URL)
        # navigate elsewhere and back to work around the third-party getting served from cache
        self.load_url(self.options_url)
        self.load_url(self.FIXTURE_URL)
        self.assert_block()

    def test_cookieblock_site_override(self):
        # verify blocking gets overriden
        self.block_domain(self.THIRD_PARTY_DOMAIN)
        self.add_site_override(self.THIRD_PARTY_DOMAIN, self.FIXTURE_DOMAIN)

        self.load_url(self.FIXTURE_URL)
        self.assert_load()

        # verify cookieblocking
        self.block_domain(self.COOKIE_DOMAIN)
        self.add_site_override(self.COOKIE_DOMAIN, self.FIXTURE_DOMAIN)

        self.load_url(self.COOKIE_FIXTURE_URL)
        self.wait_for_and_switch_to_frame("iframe[src]", timeout=1)
        self.wait_for_any_text('body')
        assert self.find_el_by_css('body').text == "cookies=0", (
            "No cookies should have been sent by the cookieblocked domain")
        # reload to verify that there are still no cookies
        self.load_url(self.COOKIE_FIXTURE_URL)
        self.wait_for_and_switch_to_frame("iframe[src]", timeout=1)
        self.wait_for_any_text('body')
        assert self.find_el_by_css('body').text == "cookies=0", (
            "No cookies should have been sent by the cookieblocked domain")

    def test_userblock_overrides_site_override(self):
        self.add_site_override(self.THIRD_PARTY_DOMAIN, self.FIXTURE_DOMAIN)
        self.set_user_action(self.THIRD_PARTY_DOMAIN, "block")
        self.load_url(self.FIXTURE_URL)
        self.assert_block()

    def test_disabling_on_site(self):
        self.block_domain(self.THIRD_PARTY_DOMAIN)
        self.disable_badger_on_site(self.FIXTURE_URL)

        self.load_url(self.FIXTURE_URL)
        self.assert_load()

    def test_disabling_on_site_parent_domain(self):
        self.block_domain(self.THIRD_PARTY_DOMAIN)
        self.disable_badger_on_site(self.FIXTURE_PARENT_DOMAIN)

        self.load_url(self.FIXTURE_URL)
        self.assert_load()

    def test_disabling_on_site_wildcard(self):
        self.block_domain(self.THIRD_PARTY_DOMAIN)
        self.disable_badger_on_site("*." + self.FIXTURE_PARENT_DOMAIN)

        self.load_url(self.FIXTURE_URL)
        self.assert_load()

    def test_reenabling_on_site(self):
        self.block_domain(self.THIRD_PARTY_DOMAIN)
        self.disable_badger_on_site(self.FIXTURE_URL)
        self.reenable_badger_on_site(self.FIXTURE_DOMAIN)

        self.load_url(self.FIXTURE_URL)
        self.assert_block()

    def test_ignoring_dnt_compliance(self):
        """We should ignore DNT compliance when DNT policy checking is off."""

        self.block_domain(self.THIRD_PARTY_DOMAIN)
        self.set_dnt(self.THIRD_PARTY_DOMAIN)

        self.load_url(self.FIXTURE_URL)
        self.assert_load()
        self.open_popup(self.FIXTURE_URL)
        assert self.get_domain_slider_state(self.THIRD_PARTY_DOMAIN) == "allow", (
            "DNT-compliant resource should be allowed")
        self.close_window_with_url(self.FIXTURE_URL)

        self.load_url(self.options_url)
        self.wait_for_script("return window.OPTIONS_INITIALIZED")
        self.find_el_by_css('a[href="#tab-general-settings"]').click()
        self.find_el_by_css('#check_dnt_policy_checkbox').click()

        self.load_url(self.FIXTURE_URL)
        self.assert_block()
        self.open_popup(self.FIXTURE_URL)
        assert self.get_domain_slider_state(self.THIRD_PARTY_DOMAIN) == "block", (
            "DNT-compliant resource should now be blocked")

    def test_reenabling_dnt_policy_checking(self):
        self.block_domain(self.THIRD_PARTY_DOMAIN)
        self.set_dnt(self.THIRD_PARTY_DOMAIN)

        # toggle EFF's DNT policy checking
        self.wait_for_script("return window.OPTIONS_INITIALIZED")
        self.find_el_by_css('a[href="#tab-general-settings"]').click()
        self.find_el_by_css('#check_dnt_policy_checkbox').click()
        self.driver.refresh()
        self.wait_for_script("return window.OPTIONS_INITIALIZED")
        self.find_el_by_css('a[href="#tab-general-settings"]').click()
        self.find_el_by_css('#check_dnt_policy_checkbox').click()

        self.load_url(self.FIXTURE_URL)
        self.assert_load()

    def test_removing_dnt(self):
        self.block_domain(self.THIRD_PARTY_DOMAIN)
        self.set_dnt(self.THIRD_PARTY_DOMAIN)

        self.load_url(self.FIXTURE_URL)
        self.assert_load()

        assert not self.check_dnt(self.THIRD_PARTY_DOMAIN), (
            "domain should not be DNT-compliant")

        self.load_url(self.FIXTURE_URL)
        if self.is_firefox_nightly():
            self.driver.refresh()
        self.assert_block()

    def test_removing_domain(self):
        # first block the domain
        self.block_domain(self.THIRD_PARTY_DOMAIN)

        # now remove it
        self.js(
            "chrome.runtime.sendMessage({"
            "  type: 'removeDomain',"
            "  domain: arguments[0]"
            "});", self.THIRD_PARTY_DOMAIN)

        # the domain should now load
        self.load_url(self.FIXTURE_URL)
        self.assert_load()


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python

import json
import unittest

import pytest

import pbtest

from functools import partial

from selenium.common.exceptions import NoSuchElementException
from selenium.webdriver.common.by import By

from pbtest import retry_until


class DntTest(pbtest.PBSeleniumTest):
    """Tests to make sure DNT policy checking works as expected."""

    def setUp(self):
        self.FIXTURE_DOMAIN = "efforg.github.io"
        self.FIXTURE_PARENT_DOMAIN = "github.io"
        self.FIXTURE_URL = (
            f"https://{self.FIXTURE_DOMAIN}/privacybadger-test-fixtures/html/")
        # TODO switch to scripting in Firefox (MV2) to remove delay
        # TODO https://github.com/EFForg/privacybadger/issues/2948
        self.FIXTURE_URL += "navigator_donottrack_delayed.html"

    def get_first_party_headers(self, url):
        self.load_url(url)

        text = self.driver.find_element(By.TAG_NAME, 'body').text

        try:
            # work around MS Edge JSON viewer garbage
            if text.startswith('1\n'):
                text = '{' + text.partition('{')[2]
            headers = json.loads(text)['headers']
        except ValueError:
            print(f"\nFailed to parse JSON from {repr(text)}")
            return None

        return headers

    def set_dnt_hashes(self):
        # MEGAHACK: make sha1 of "cookies=0" a valid DNT hash
        # so that the DNT policy checks to the domain that replies with cookies=X
        # will succeed when we don't send cookies along with the request
        self.load_url(self.options_url)
        self.driver.execute_async_script(
            "let done = arguments[arguments.length - 1];"
            "chrome.runtime.sendMessage({"
            "  type: 'setDntHashes',"
            "  value: { 'cookies=0 test policy': 'f63ee614ebd77f8634b92633c6bb809a64b9a3d7' }"
            "}, done);")

    def assert_navigator_gpc_unset(self, msg=""):
        # GPC on Navigator should be unset (Chrome) or False (Firefox)
        assert self.js("""
return (typeof navigator.globalPrivacyControl == 'undefined' ||
  navigator.globalPrivacyControl === false);"""), msg

    def test_dnt_policy_check_should_happen_for_blocked_domains(self):
        PAGE_URL = (
            "https://efforg.github.io/privacybadger-test-fixtures/html/"
            "recording_nontracking_domains.html"
        )
        DNT_DOMAIN = "dnt-request-cookies-test.trackersimulator.org"

        self.clear_tracker_data()

        self.set_dnt_hashes()

        # mark the "DNT-compliant" domain for blocking
        self.block_domain(DNT_DOMAIN)

        # visit a page that loads a resource from that domain
        self.load_url(PAGE_URL)

        # verify that the domain is blocked
        self.open_popup(PAGE_URL)
        assert self.get_domain_slider_state(DNT_DOMAIN) == "block", (
            "DNT-compliant resource should have been blocked at first")

        def reload_and_see_if_unblocked():
            # switch back to the page with the DNT-compliant resource
            self.switch_to_window_with_url(PAGE_URL)

            # reload it
            self.load_url(PAGE_URL)

            self.open_popup(PAGE_URL)
            return self.get_domain_slider_state(DNT_DOMAIN) == "block"

        # verify that the domain is allowed
        was_blocked = retry_until(
            reload_and_see_if_unblocked,
            tester=lambda x: not x,
            msg="Waiting a bit for DNT check to complete and retrying ...")

        assert not was_blocked, (
            "DNT-compliant resource should have gotten unblocked")

    def test_dnt_policy_check_should_not_set_cookies(self):
        TEST_DOMAIN = "dnt-test.trackersimulator.org"
        TEST_URL = f"https://{TEST_DOMAIN}/"

        # verify that the domain itself doesn't set cookies
        self.load_url(TEST_URL)
        assert not self.driver.get_cookies(), "Expect no cookies at first"

        # directly visit a DNT policy URL known to set cookies
        self.load_url(TEST_URL + ".well-known/dnt-policy.txt")
        assert len(self.driver.get_cookies()) == 1, (
            "DNT policy URL should have set a cookie")

        # verify we got a cookie
        self.load_url(TEST_URL)
        assert len(self.driver.get_cookies()) == 1, "Should still have one cookie"

        # clear cookies and verify
        self.driver.delete_all_cookies()
        self.load_url(TEST_URL)
        assert not self.driver.get_cookies(), "Should have no cookies again"

        # perform a DNT policy check
        self.check_dnt(TEST_DOMAIN)

        # check that we didn't get cookied by the DNT URL
        self.load_url(TEST_URL)
        assert not self.driver.get_cookies(), (
            "Shouldn't have any cookies after the DNT check")

    def test_dnt_policy_check_should_not_send_cookies(self):
        TEST_DOMAIN = "dnt-request-cookies-test.trackersimulator.org"
        TEST_URL = f"https://{TEST_DOMAIN}/"

        # directly visit a DNT policy URL known to set cookies
        self.load_url(TEST_URL + ".well-known/dnt-policy.txt")
        assert len(self.driver.get_cookies()) == 1, (
            "DNT policy URL should have set a cookie")

        # how to check we didn't send a cookie along with request?
        # the DNT policy URL used by this test returns "cookies=X"
        # where X is the number of cookies it got
        self.set_dnt_hashes()

        # perform a DNT policy check
        result = self.check_dnt(TEST_DOMAIN)
        assert result, "One or more cookies were sent (cookies=0 policy hash did not match)"

    @pytest.mark.flaky(reruns=3, condition=pbtest.shim.browser_type in ("chrome", "edge"))
    def test_should_not_record_nontracking_domains(self):
        NONTRACKING_FIXTURE_URL = (
            "https://efforg.github.io/privacybadger-test-fixtures/html/"
            "recording_nontracking_domains.html"
        )
        TRACKING_DOMAIN = "dnt-request-cookies-test.trackersimulator.org"
        NON_TRACKING_DOMAIN = "www.eff.org"

        # clear pre-trained/seed tracker data
        self.clear_tracker_data()

        # enable local learning
        self.wait_for_script("return window.OPTIONS_INITIALIZED")
        self.find_el_by_css('a[href="#tab-general-settings"]').click()
        self.find_el_by_css('#local-learning-checkbox').click()

        # visit a page containing two third-party resources,
        # one from a cookie-tracking domain
        # and one from a non-tracking domain
        self.load_url(NONTRACKING_FIXTURE_URL)

        # verify both domains are present on the page
        try:
            selector = f"iframe[src*='{TRACKING_DOMAIN}']"
            self.driver.find_element(By.CSS_SELECTOR, selector)
        except NoSuchElementException:
            self.fail("Unable to find the tracking domain on the page")
        try:
            selector = f"img[src*='{NON_TRACKING_DOMAIN}']"
            self.driver.find_element(By.CSS_SELECTOR, selector)
        except NoSuchElementException:
            self.fail("Unable to find the non-tracking domain on the page")

        action_map = self.get_badger_storage('action_map')

        # verify that the cookie-tracking domain was recorded
        assert TRACKING_DOMAIN in action_map, (
            "Tracking domain should have gotten recorded")

        # verify that the non-tracking domain was not recorded
        assert NON_TRACKING_DOMAIN not in action_map, (
            "Non-tracking domain should not have gotten recorded")

    def test_first_party_dnt_header(self):
        TEST_URL = "https://httpbingo.org/get"
        headers = retry_until(partial(self.get_first_party_headers, TEST_URL),
                              times=8)
        assert headers is not None, "It seems we failed to get headers"
        assert 'Dnt' in headers, "DNT header should have been present"
        assert 'Sec-Gpc' in headers, "GPC header should have been present"
        assert headers['Dnt'] == ["1"], 'DNT header should have been set to "1"'
        assert headers['Sec-Gpc'] == ["1"], 'Sec-Gpc header should have been set to "1"'

    def test_no_dnt_header_when_disabled_on_site(self):
        TEST_URL = "https://httpbingo.org/get"
        self.disable_badger_on_site(TEST_URL)
        headers = retry_until(partial(self.get_first_party_headers, TEST_URL),
                              times=8)
        assert headers is not None, "It seems we failed to get headers"
        assert 'Dnt' not in headers, "DNT header should have been missing"
        assert 'Sec-Gpc' not in headers, "GPC header should have been missing"

    def test_no_dnt_header_when_dnt_disabled(self):
        TEST_URL = "https://httpbingo.org/get"

        self.load_url(self.options_url)
        self.wait_for_script("return window.OPTIONS_INITIALIZED")
        self.find_el_by_css('a[href="#tab-general-settings"]').click()
        self.find_el_by_css('#enable_dnt_checkbox').click()

        headers = retry_until(partial(self.get_first_party_headers, TEST_URL),
                              times=8)
        assert headers is not None, "It seems we failed to get headers"
        assert 'Dnt' not in headers, "DNT header should have been missing"
        assert 'Sec-Gpc' not in headers, "GPC header should have been missing"

    def test_navigator_object(self):
        self.load_url(self.FIXTURE_URL, wait_for_body_text=True)
        body_text = self.driver.find_element(By.TAG_NAME, 'body').text
        assert body_text == 'no tracking (navigator.doNotTrack="1")', (
            'navigator.doNotTrack should have been set to "1"')
        assert self.js("return navigator.globalPrivacyControl === true"), (
            "navigator.globalPrivacyControl should have been set to true")
        assert self.js("""
return Object.getOwnPropertyDescriptor(
  Navigator.prototype, 'globalPrivacyControl')?.get?.call(navigator);
"""), "GPC should be set on Navigator.prototype"

    def test_navigator_unmodified_when_disabled_on_site(self):
        self.disable_badger_on_site(self.FIXTURE_URL)

        self.load_url(self.FIXTURE_URL, wait_for_body_text=True)

        # navigator.doNotTrack defaults to null in Chrome, "unspecified" in Firefox
        body_text = self.driver.find_element(By.TAG_NAME, 'body').text
        assert body_text[0:5] == 'unset', (
            "navigator.doNotTrack should be unset or \"unspecified\"")

        self.assert_navigator_gpc_unset("navigator.globalPrivacyControl should be unset or False")

    def test_navigator_disabling_on_site_parent_domain(self):
        """Needs to be consistent with test_disabling_on_site_parent_domain()"""
        self.disable_badger_on_site(self.FIXTURE_PARENT_DOMAIN)
        self.load_url(self.FIXTURE_URL, wait_for_body_text=True)
        self.assert_navigator_gpc_unset("navigator.globalPrivacyControl should be unset or False")

    def test_navigator_disabling_on_site_wildcard(self):
        """Needs to be consistent with test_disabling_on_site_wildcard()"""
        self.disable_badger_on_site("*." + self.FIXTURE_PARENT_DOMAIN)
        self.load_url(self.FIXTURE_URL, wait_for_body_text=True)
        self.assert_navigator_gpc_unset("navigator.globalPrivacyControl should be unset or False")

    def test_navigator_unmodified_when_dnt_disabled(self):
        self.load_url(self.options_url)
        self.wait_for_script("return window.OPTIONS_INITIALIZED")
        self.find_el_by_css('a[href="#tab-general-settings"]').click()
        self.find_el_by_css('#enable_dnt_checkbox').click()

        self.load_url(self.FIXTURE_URL, wait_for_body_text=True)

        # navigator.doNotTrack defaults to null in Chrome, "unspecified" in Firefox
        body_text = self.driver.find_element(By.TAG_NAME, 'body').text
        assert body_text[0:5] == 'unset', (
            "navigator.doNotTrack should be unset or \"unspecified\"")

        self.assert_navigator_gpc_unset("navigator.globalPrivacyControl should be unset or False")

    def test_navigator_toggling_dnt_and_disabled_sites(self):
        # disable on site
        self.disable_badger_on_site(self.FIXTURE_URL)

        # disable sending DNT signals
        self.load_url(self.options_url)
        self.wait_for_script("return window.OPTIONS_INITIALIZED")
        self.find_el_by_css('a[href="#tab-general-settings"]').click()
        self.find_el_by_css('#enable_dnt_checkbox').click()

        self.load_url(self.FIXTURE_URL, wait_for_body_text=True)
        assert self.js("return navigator.doNotTrack") != "1", (
            "navigator.doNotTrack should not be set")
        self.assert_navigator_gpc_unset("navigator.globalPrivacyControl should be unset or False")

        # re-enable sending DNT signals
        self.load_url(self.options_url)
        self.wait_for_script("return window.OPTIONS_INITIALIZED")
        self.find_el_by_css('a[href="#tab-general-settings"]').click()
        self.find_el_by_css('#enable_dnt_checkbox').click()

        self.load_url(self.FIXTURE_URL, wait_for_body_text=True)
        assert self.js("return navigator.doNotTrack") != "1", (
            "navigator.doNotTrack should still not be set")
        self.assert_navigator_gpc_unset("navigator.globalPrivacyControl should still be unset or False")

        # re-enable on site
        self.reenable_badger_on_site("efforg.github.io")

        self.load_url(self.FIXTURE_URL, wait_for_body_text=True)
        assert self.js("return navigator.doNotTrack") == "1", (
            "navigator.doNotTrack should now be set")
        assert self.js("return navigator.globalPrivacyControl"), (
            "navigator.globalPrivacyControl should also be set")


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python

import unittest

import pbtest

from time import sleep

from selenium.webdriver.common.by import By
from selenium.common.exceptions import (
    NoSuchElementException,
    StaleElementReferenceException,
    TimeoutException
)
from selenium.webdriver.common.keys import Keys
from selenium.webdriver.support.ui import Select


class WidgetsTest(pbtest.PBSeleniumTest):

    FIXTURES_URL = "https://efforg.github.io/privacybadger-test-fixtures/html/"
    FIXTURES_HOST = "efforg.github.io"
    BASIC_FIXTURE_URL = FIXTURES_URL + "widget_basic.html"
    DYNAMIC_FIXTURE_URL = FIXTURES_URL + "widget_dynamic.html"
    THIRD_PARTY_DOMAIN = "privacybadger-tests.eff.org"
    TYPE3_WIDGET_NAME = "Type 3 Widget"
    TYPE4_WIDGET_NAME = "Type 4 Widget"
    TYPE4_WIDGET_CLASS = "pb-type4-test-widget"

    def setUp(self):
        self.set_up_widgets()

    def set_up_widgets(self):
        """Reinitializes Privacy Badger's widget replacement definitions."""

        widgetsJson = {
            self.TYPE3_WIDGET_NAME: {
                "domain": self.THIRD_PARTY_DOMAIN,
                "buttonSelectors": [
                    "iframe#pb-type3-test-widget"
                ],
                "replacementButton": {
                    "unblockDomains": [
                        self.THIRD_PARTY_DOMAIN
                    ],
                    "type": 3
                }
            },
            self.TYPE4_WIDGET_NAME: {
                "domains": [
                    self.THIRD_PARTY_DOMAIN
                ],
                "buttonSelectors": [
                    "div." + self.TYPE4_WIDGET_CLASS
                ],
                "scriptSelectors": [
                    "script." + self.TYPE4_WIDGET_CLASS
                ],
                "replacementButton": {
                    "unblockDomains": [
                        self.THIRD_PARTY_DOMAIN
                    ],
                    "type": 4
                }
            }
        }

        # reinitialize widgets using above JSON
        self.load_url(self.options_url)
        self.driver.execute_async_script((
            "let done = arguments[arguments.length - 1];"
            "(async function (widgetsJson) {"
            "  const { default: widgetLoader } = await import('../js/socialwidgetloader.js');"
            "  chrome.runtime.sendMessage({"
            "    type: 'setWidgetList',"
            "    value: widgetLoader.initializeWidgets(widgetsJson)"
            "  }, done);"
            "}(arguments[0]));"
        ), widgetsJson)

    def switch_to_frame(self, selector):
        self.wait_for_and_switch_to_frame(selector, timeout=1)

    def assert_widget(self, kind="type3"):
        if kind == "type3":
            self._assert_type3_widget()
        elif kind == "type4":
            self._assert_type4_widget()
        else:
            self.fail("Unknown widget type")

    def _assert_type3_widget(self):
        try:
            self.switch_to_frame('iframe[src]')
        except (StaleElementReferenceException, TimeoutException):
            self.fail("Unable to find widget frame")

        try:
            self.wait_for_text('body', "Hello world!")
        except TimeoutException:
            self.fail("Unable to find expected widget text")

        self.driver.switch_to.default_content()

    def _assert_type4_widget(self):
        try:
            self.wait_for_text('div.' + self.TYPE4_WIDGET_CLASS,
                "A third-party widget script was here")
        except TimeoutException:
            self.fail("Unable to find expected widget output")

    def assert_replacement(self, widget_name=None):
        if not widget_name:
            widget_name = self.TYPE3_WIDGET_NAME

        try:
            self.switch_to_frame(f'iframe[srcdoc*="{widget_name}"]')
        except (StaleElementReferenceException, TimeoutException):
            self.fail("Unable to find widget placeholder frame")

        try:
            self.find_el_by_css("button[id^='btn-once-']")
            self.find_el_by_css("button[id^='btn-site-']")
        except TimeoutException:
            self.fail("Unable to find expected widget placeholder buttons")

        self.driver.switch_to.default_content()

    def assert_widget_blocked(self):
        try:
            self.switch_to_frame('iframe[src]')
        except TimeoutException:
            self.fail("Widget frame should still be here")

        frame_text = self.txt_by_css('body')
        # TODO doesn't work in non-English locales
        if frame_text != "This page has been blocked by an extension":
            assert not frame_text, "Widget frame should be empty"

        self.driver.switch_to.default_content()

    def assert_no_widget(self):
        try:
            self.switch_to_frame('iframe[src]')
            self.fail("Widget frame should be missing")
        except TimeoutException:
            pass
        self.driver.switch_to.default_content()

    def assert_no_replacement(self, widget_name=None):
        if not widget_name:
            widget_name = self.TYPE3_WIDGET_NAME
        try:
            self.switch_to_frame(f'iframe[srcdoc*="{widget_name}"]')
            self.fail("Widget placeholder frame should be missing")
        except TimeoutException:
            pass
        self.driver.switch_to.default_content()

    def activate_widget(self, widget_name=None, once=True):
        if not widget_name:
            widget_name = self.TYPE3_WIDGET_NAME
        id_prefix = 'btn-once' if once else 'btn-site'
        self.switch_to_frame(f'iframe[srcdoc*="{widget_name}"]')
        self.find_el_by_css(f"button[id^='{id_prefix}']").click()
        self.driver.switch_to.default_content()
        # wait a bit for the widget to get reinserted
        sleep(1)

    def test_replacement_basic(self):
        # visit the basic widget fixture
        self.load_url(self.BASIC_FIXTURE_URL)
        # verify the widget is present
        self.assert_widget()

        # block the test widget's domain
        self.block_domain(self.THIRD_PARTY_DOMAIN)

        # revisit the fixture
        self.load_url(self.BASIC_FIXTURE_URL)
        # verify the widget got replaced
        self.assert_replacement()

    def test_replacement_dynamic(self):
        # visit the dynamic widget fixture
        self.load_url(self.DYNAMIC_FIXTURE_URL)
        # verify the widget is initially missing
        self.assert_no_widget()

        # verify the widget shows up once you click on the trigger element
        self.find_el_by_css('#widget-trigger').click()
        self.assert_widget()

        # block the test widget's domain
        self.block_domain(self.THIRD_PARTY_DOMAIN)

        # revisit the fixture
        self.load_url(self.DYNAMIC_FIXTURE_URL)
        # click on the trigger element
        self.find_el_by_css('#widget-trigger').click()
        # verify the widget got replaced
        self.assert_replacement()

    def test_activation(self):
        self.block_domain(self.THIRD_PARTY_DOMAIN)
        self.load_url(self.BASIC_FIXTURE_URL)
        self.assert_replacement()

        # click the "allow once" button
        self.activate_widget()

        # verify the original widget is restored
        self.assert_widget()

        # verify the type 4 widget is still replaced
        try:
            self.driver.find_element(By.CSS_SELECTOR,
                'div.' + self.TYPE4_WIDGET_CLASS)
            self.fail("Widget output container div should be missing")
        except NoSuchElementException:
            pass
        self.assert_replacement(self.TYPE4_WIDGET_NAME)

        self.activate_widget(self.TYPE4_WIDGET_NAME)

        # assert all script attributes were copied
        script_el = self.driver.find_element(By.CSS_SELECTOR,
            'script.' + self.TYPE4_WIDGET_CLASS)
        assert script_el.get_dom_attribute('async') == "true"
        assert script_el.get_dom_attribute('data-foo') == "bar"

        self.assert_widget("type4")

    def test_activation_site(self):
        self.block_domain(self.THIRD_PARTY_DOMAIN)
        self.load_url(self.BASIC_FIXTURE_URL)
        self.assert_replacement()

        # click the "allow once" button
        self.activate_widget()

        # verify the original widget is restored
        self.assert_widget()

        # open a new window (to get around widget activation caching)
        self.open_window()
        self.load_url(self.BASIC_FIXTURE_URL)

        # verify the widget got replaced
        self.assert_replacement()

        # click the "allow on site" button
        self.activate_widget(once=False)

        # verify the original widget is restored
        self.assert_widget()

        # open a new window (to get around widget activation caching)
        self.open_window()
        self.load_url(self.BASIC_FIXTURE_URL)

        # verify basic widget is neither replaced nor blocked
        self.assert_no_replacement()
        self.assert_widget()

        # remove the site exception
        self.load_url(self.options_url)
        self.wait_for_script("return window.OPTIONS_INITIALIZED")
        self.find_el_by_css('a[href="#tab-manage-widgets"]').click()
        select = Select(self.driver.find_element(By.ID, 'widget-site-exceptions-select'))
        select.select_by_value(self.FIXTURES_HOST)
        self.driver.find_element(By.ID, 'widget-site-exceptions-remove-button').click()

        # verify basic widget is replaced again
        self.open_window()
        self.load_url(self.BASIC_FIXTURE_URL)
        self.assert_replacement()

    def test_disabling_site(self):
        self.block_domain(self.THIRD_PARTY_DOMAIN)

        self.disable_badger_on_site(self.BASIC_FIXTURE_URL)

        # verify basic widget is neither replaced nor blocked
        self.load_url(self.BASIC_FIXTURE_URL)
        self.assert_no_replacement()
        self.assert_widget()
        # type 4 replacement should also be missing
        self.assert_no_replacement(self.TYPE4_WIDGET_NAME)
        # while the type 4 widget script should have executed
        self.assert_widget("type4")

        # verify dynamic widget is neither replaced nor blocked
        self.load_url(self.DYNAMIC_FIXTURE_URL)
        self.find_el_by_css('#widget-trigger').click()
        self.assert_no_replacement()
        self.assert_widget()

    def test_disabling_replacement_for_one_widget(self):
        self.block_domain(self.THIRD_PARTY_DOMAIN)

        # add the widget to the list of exceptions
        self.load_url(self.options_url)
        self.wait_for_script("return window.OPTIONS_INITIALIZED")
        self.find_el_by_css('a[href="#tab-manage-widgets"]').click()
        self.find_el_by_css('input[type="search"]').send_keys(
            self.TYPE3_WIDGET_NAME, Keys.ENTER)

        # verify basic widget is no longer replaced
        self.load_url(self.BASIC_FIXTURE_URL)
        self.assert_no_replacement()
        self.assert_widget_blocked()
        # verify the type 4 widget is still replaced
        self.assert_replacement(self.TYPE4_WIDGET_NAME)

        # verify dynamic widget is no longer replaced
        self.load_url(self.DYNAMIC_FIXTURE_URL)
        self.find_el_by_css('#widget-trigger').click()
        self.assert_no_replacement()
        self.assert_widget_blocked()

    def test_no_replacement_when_cookieblocked(self):
        self.cookieblock_domain(self.THIRD_PARTY_DOMAIN)
        self.load_url(self.BASIC_FIXTURE_URL)

        self.assert_no_replacement()
        self.assert_no_replacement(self.TYPE4_WIDGET_NAME)

        self.assert_widget()
        self.assert_widget("type4")


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3

import argparse
import json

from collections import OrderedDict

def update_manifest(manifest_path, operation, key, value, json_value):
    with open(manifest_path, 'r', encoding="utf-8") as f:
        manifest_data = json.load(f, object_pairs_hook=OrderedDict)

    if operation == 'del':
        data = manifest_data
        while '.' in key: # handle nested keys ("x.y.z")
            idx, _, key = key.partition('.')
            data = data[idx]
        del data[key]

    elif operation == 'set':
        if json_value:
            try:
                value = json.loads(value)
            except json.decoder.JSONDecodeError as err:
                print("Failed to parse", value)
                raise err

        data = manifest_data
        while '.' in key: # handle nested keys ("x.y.z")
            idx, _, key = key.partition('.')
            if idx not in data:
                data[idx] = {}
            data = data[idx]
        data[key] = value

    with open(manifest_path, 'w', encoding="utf-8") as f:
        json.dump(manifest_data, f, sort_keys=False, indent=2, separators=(',', ': '))
        f.write('\n')


if __name__ == '__main__':
    parser = argparse.ArgumentParser()

    parser.add_argument('manifest_path')
    parser.add_argument('operation', choices=('set', 'del'))
    parser.add_argument('key')
    parser.add_argument('value', nargs='?')
    parser.add_argument('--json-value', action='store_true', default=False)

    args = parser.parse_args()

    update_manifest(args.manifest_path, args.operation, args.key, args.value, args.json_value)
